from abc import abstractmethod


class IMessageHandler:
    """

    """
    def __init__(self):
        self.initialed = False
        self.params = {}

    def is_initialed(self)->bool:
        return self.initialed

    def set_initialed(self, initialed=True):
        self.initialed = initialed

    def set_params(self,params):
        self.params = params

    def check_ficus_online(self):
        """
        检查ficus服务是否存在
        :param eureka_client:
        :return:
        """
        from client import check_instance_avaliable
        from api.exceptions import ServiceNoInstanceException
        try:
            check_instance_avaliable()
            return True
        except ServiceNoInstanceException:
            return False

    @abstractmethod
    def message(self,data):
        """
        将消息处理后存到目标fd,需要实现
        :param data:
        :return:
        """
        pass

    @abstractmethod
    def kill(self):
        """
        需要实现的结束的东西
        :return:
        """
        pass


# -*- coding: utf-8 -*-

import atexit
import json
import os
import re
import socket
import time
import random
import inspect
import xml.etree.ElementTree as ElementTree
from threading import Timer
from threading import RLock
from threading import Thread
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

from py_eureka_client.__logger__ import get_logger
import py_eureka_client.http_client as http_client

try:
    long(0)
except NameError:
    # python 3 does no longer support long method, use int instead
    long = int

_logger = get_logger("EurekaClient")

"""
Status of instances
"""
INSTANCE_STATUS_UP = "UP"
INSTANCE_STATUS_DOWN = "DOWN"
INSTANCE_STATUS_STARTING = "STARTING"
INSTANCE_STATUS_OUT_OF_SERVICE = "OUT_OF_SERVICE"
INSTANCE_STATUS_UNKNOWN = "UNKNOWN"

"""
Action type of instances
"""
ACTION_TYPE_ADDED = "ADDED"
ACTION_TYPE_MODIFIED = "MODIFIED"
ACTION_TYPE_DELETED = "DELETED"

""" 
This is for the DiscoveryClient, when this strategy is set, get_service_url will random choose one of the UP instance and return its url
This is the default strategy
"""
HA_STRATEGY_RANDOM = 1
"""
This is for the DiscoveryClient, when this strategy is set, get_service_url will always return one instance until it is down
"""
HA_STRATEGY_STICK = 2
"""
This is for the DiscoveryClient, when this strategy is set, get_service_url will always return a new instance if any other instances are up
"""
HA_STRATEGY_OTHER = 3

"""
The timeout seconds that all http request to the eureka server
"""
_DEFAULT_TIME_OUT = 5
"""
Default eureka server url.
"""
_DEFAULT_EUREKA_SERVER_URL = "http://127.0.0.1:8761/eureka/"
"""
Default instance field values
"""
_DEFAULT_INSTNACE_PORT = 9090
_DEFAULT_INSTNACE_SECURE_PORT = 9443
_RENEWAL_INTERVAL_IN_SECS = 30
_DURATION_IN_SECS = 90
_DEFAULT_DATA_CENTER_INFO = "MyOwn"
_DEFAULT_DATA_CENTER_INFO_CLASS = "com.netflix.appinfo.InstanceInfo$DefaultDataCenterInfo"
"""
Default encoding
"""
_DEFAULT_ENCODING = "utf-8"

### =========================> Base Mehods <======================================== ###
### Beans ###


class Applications:

    def __init__(self,
                 apps__hashcode="",
                 versions__delta="",
                 applications=None):
        self.apps__hashcode = apps__hashcode
        self.versions__delta = versions__delta
        self.__applications = applications if applications is not None else []
        self.__application_name_dic = {}
        self.__app_lock = RLock()

    @property
    def appsHashcode(self):
        return self.apps__hashcode

    @property
    def applications(self):
        return self.__applications

    @property
    def versionsDelta(self):
        return self.versions__delta

    def add_application(self, application):
        with self.__app_lock:
            self.__applications.append(application)
            self.__application_name_dic[application.name] = application

    def get_application(self, app_name):
        with self.__app_lock:
            if app_name in self.__application_name_dic:
                return self.__application_name_dic[app_name]
            else:
                return Application(name=app_name)


class Application:

    def __init__(self,
                 name="",
                 instances=None):
        self.name = name
        self.__instances = instances if instances is not None else []
        self.__instances_dict = {}
        self.__inst_lock = RLock()

    @property
    def instances(self):
        with self.__inst_lock:
            return self.__instances

    @property
    def up_instances(self):
        with self.__inst_lock:
            up_inst = []
            for item in self.__instances:
                if item.status == INSTANCE_STATUS_UP:
                    up_inst.append(item)
            return up_inst

    def get_instance(self, instance_id):
        with self.__inst_lock:
            if instance_id in self.__instances_dict:
                return self.__instances_dict[instance_id]
            else:
                return None

    def add_instance(self, instance):
        with self.__inst_lock:
            self.__instances.append(instance)
            self.__instances_dict[instance.instanceId] = instance

    def update_instance(self, instance):
        with self.__inst_lock:
            _logger.debug("update instance %s" % instance.instanceId)
            updated = False
            for idx in range(len(self.__instances)):
                ele = self.__instances[idx]
                if ele.instanceId == instance.instanceId:
                    _logger.debug("updating index %d" % idx)
                    self.__instances[idx] = instance
                    updated = True
                    break

            if not updated:
                self.add_instance(instance)

    def remove_instance(self, instance):
        with self.__inst_lock:
            for idx in range(len(self.__instances)):
                ele = self.__instances[idx]
                if ele.instanceId == instance.instanceId:
                    del self.__instances[idx]
                    break
            if instance.instanceId in self.__instances_dict:
                del self.__instances_dict[instance.instanceId]


class LeaseInfo:

    def __init__(self,
                 renewalIntervalInSecs=_RENEWAL_INTERVAL_IN_SECS,
                 durationInSecs=_DURATION_IN_SECS,
                 registrationTimestamp=0,
                 lastRenewalTimestamp=0,
                 renewalTimestamp=0,
                 evictionTimestamp=0,
                 serviceUpTimestamp=0):
        self.renewalIntervalInSecs = renewalIntervalInSecs
        self.durationInSecs = durationInSecs
        self.registrationTimestamp = registrationTimestamp
        self.lastRenewalTimestamp = lastRenewalTimestamp
        self.renewalTimestamp = renewalTimestamp
        self.evictionTimestamp = evictionTimestamp
        self.serviceUpTimestamp = serviceUpTimestamp


class DataCenterInfo:

    def __init__(self,
                 name=_DEFAULT_DATA_CENTER_INFO,  # Netflix, Amazon, MyOwn
                 className=_DEFAULT_DATA_CENTER_INFO_CLASS):
        self.name = name
        self.className = className


class PortWrapper:
    def __init__(self, port=0, enabled=False):
        self.port = port
        self.enabled = enabled


class Instance:

    def __init__(self,
                 instanceId="",
                 sid="",  # @deprecated
                 app="",
                 appGroupName="",
                 ipAddr="",
                 port=PortWrapper(port=_DEFAULT_INSTNACE_PORT, enabled=True),
                 securePort=PortWrapper(port=_DEFAULT_INSTNACE_SECURE_PORT, enabled=False),
                 homePageUrl="",
                 statusPageUrl="",
                 healthCheckUrl="",
                 secureHealthCheckUrl="",
                 vipAddress="",
                 secureVipAddress="",
                 countryId=1,
                 dataCenterInfo=DataCenterInfo(),
                 hostName="",
                 status="",  # UP, DOWN, STARTING, OUT_OF_SERVICE, UNKNOWN
                 overriddenstatus="",  # UP, DOWN, STARTING, OUT_OF_SERVICE, UNKNOWN
                 leaseInfo=LeaseInfo(),
                 isCoordinatingDiscoveryServer=False,
                 metadata=None,
                 lastUpdatedTimestamp=0,
                 lastDirtyTimestamp=0,
                 actionType=ACTION_TYPE_ADDED,  # ADDED, MODIFIED, DELETED
                 asgName=""):
        self.instanceId = instanceId
        self.sid = sid
        self.app = app
        self.appGroupName = appGroupName
        self.ipAddr = ipAddr
        self.port = port
        self.securePort = securePort
        self.homePageUrl = homePageUrl
        self.statusPageUrl = statusPageUrl
        self.healthCheckUrl = healthCheckUrl
        self.secureHealthCheckUrl = secureHealthCheckUrl
        self.vipAddress = vipAddress
        self.secureVipAddress = secureVipAddress
        self.countryId = countryId
        self.dataCenterInfo = dataCenterInfo
        self.hostName = hostName
        self.status = status
        self.overriddenstatus = overriddenstatus
        self.leaseInfo = leaseInfo
        self.isCoordinatingDiscoveryServer = isCoordinatingDiscoveryServer
        self.metadata = metadata if metadata is not None else {}
        self.lastUpdatedTimestamp = lastUpdatedTimestamp
        self.lastDirtyTimestamp = lastDirtyTimestamp
        self.actionType = actionType
        self.asgName = asgName


########################## Basic functions #################################
####### Registry functions #########
def register(eureka_server, instance):
    instance_dic = {
        'instanceId': instance.instanceId,
        'hostName': instance.hostName,
        'app': instance.app,
        'ipAddr': instance.ipAddr,
        'status': instance.status,
        'overriddenstatus': instance.overriddenstatus,
        'port': {
            '$': instance.port.port,
            '@enabled': str(instance.port.enabled).lower()
        },
        'securePort': {
            '$': instance.securePort.port,
            '@enabled': str(instance.securePort.enabled).lower()
        },
        'countryId': instance.countryId,
        'dataCenterInfo': {
            '@class': instance.dataCenterInfo.className,
            'name': instance.dataCenterInfo.name
        },
        'leaseInfo': {
            'renewalIntervalInSecs': instance.leaseInfo.renewalIntervalInSecs,
            'durationInSecs': instance.leaseInfo.durationInSecs,
            'registrationTimestamp': instance.leaseInfo.registrationTimestamp,
            'lastRenewalTimestamp': instance.leaseInfo.lastRenewalTimestamp,
            'evictionTimestamp': instance.leaseInfo.evictionTimestamp,
            'serviceUpTimestamp': instance.leaseInfo.serviceUpTimestamp
        },
        'metadata': instance.metadata,
        'homePageUrl': instance.homePageUrl,
        'statusPageUrl': instance.statusPageUrl,
        'healthCheckUrl': instance.healthCheckUrl,
        'secureHealthCheckUrl': instance.secureHealthCheckUrl,
        'vipAddress': instance.vipAddress,
        'secureVipAddress': instance.secureVipAddress,
        'lastUpdatedTimestamp': str(instance.lastUpdatedTimestamp),
        'lastDirtyTimestamp': str(instance.lastDirtyTimestamp),
        'isCoordinatingDiscoveryServer': str(instance.isCoordinatingDiscoveryServer).lower()
    }
    _register(eureka_server, instance_dic)


def _register(eureka_server, instance_dic):
    req = http_client.Request(_format_url(eureka_server) + "apps/%s" % instance_dic["app"])
    req.add_header('Content-Type', 'application/json')
    req.get_method = lambda: "POST"
    http_client.load(req, json.dumps({"instance": instance_dic}).encode(_DEFAULT_ENCODING), timeout=_DEFAULT_TIME_OUT)


def cancel(eureka_server, app_name, instance_id):
    req = http_client.Request(_format_url(eureka_server) + "apps/%s/%s" % (app_name, instance_id))
    req.get_method = lambda: "DELETE"
    http_client.load(req, timeout=_DEFAULT_TIME_OUT)


def send_heart_beat(eureka_server, app_name, instance_id, last_dirty_timestamp, status=INSTANCE_STATUS_UP, overriddenstatus=""):
    url = _format_url(eureka_server) + "apps/%s/%s?status=%s&lastDirtyTimestamp=%s" % \
        (app_name, instance_id, status, str(last_dirty_timestamp))
    _logger.debug("heartbeat url::" + url)
    if overriddenstatus != "":
        url += "&overriddenstatus=" + overriddenstatus

    req = http_client.Request(url)
    req.get_method = lambda: "PUT"
    http_client.load(req, timeout=_DEFAULT_TIME_OUT)


def status_update(eureka_server, app_name, instance_id, last_dirty_timestamp, status):
    url = _format_url(eureka_server) + "apps/%s/%s?status=%s&lastDirtyTimestamp=%s" % \
        (app_name, instance_id, status, str(last_dirty_timestamp))

    req = http_client.Request(url)
    req.get_method = lambda: "PUT"
    http_client.load(req, timeout=_DEFAULT_TIME_OUT)


def delete_status_override(eureka_server, app_name, instance_id, last_dirty_timestamp):
    url = _format_url(eureka_server) + "apps/%s/%s/status?lastDirtyTimestamp=%s" % \
        (app_name, instance_id, str(last_dirty_timestamp))

    req = http_client.Request(url)
    req.get_method = lambda: "DELETE"
    http_client.load(req, timeout=_DEFAULT_TIME_OUT)


####### Discovory functions ########


def get_applications(eureka_server, regions=[]):
    return _get_applications_(_format_url(eureka_server) + "apps/", regions)


def _format_url(url):
    if url.endswith('/'):
        return url
    else:
        return url + "/"


def _get_applications_(url, regions=[]):
    _url = url
    if len(regions) > 0:
        _url = _url + ("&" if "?" in _url else "?") + "regions=" + (",".join(regions))

    txt = http_client.load(_url, timeout=_DEFAULT_TIME_OUT)
    return _build_applications(ElementTree.fromstring(txt))


def _build_applications(xml_node):
    if xml_node.tag != "applications":
        return None
    applications = Applications()
    for child_node in list(xml_node):
        if child_node.tag == "versions__delta" and child_node.text is not None:
            applications.versions__delta = child_node.text
        elif child_node.tag == "apps__hashcode" and child_node.text is not None:
            applications.apps__hashcode = child_node.text
        elif child_node.tag == "application":
            applications.add_application(_build_application(child_node))

    return applications


def _build_application(xml_node):
    if xml_node.tag != "application":
        return None
    application = Application()
    for child_node in xml_node:
        if child_node.tag == "name":
            application.name = child_node.text
        elif child_node.tag == "instance":
            application.add_instance(_build_instance(child_node))
    return application


def _build_instance(xml_node):
    if xml_node.tag != "instance":
        return None
    instance = Instance()
    for child_node in xml_node:
        if child_node.tag == "instanceId":
            instance.instanceId = child_node.text
        elif child_node.tag == "sid":
            instance.sid = child_node.text
        elif child_node.tag == "app":
            instance.app = child_node.text
        elif child_node.tag == "appGroupName":
            instance.appGroupName = child_node.text
        elif child_node.tag == "ipAddr":
            instance.ipAddr = child_node.text
        elif child_node.tag == "port":
            instance.port = _build_port(child_node)
        elif child_node.tag == "securePort":
            instance.securePort = _build_port(child_node)
        elif child_node.tag == "homePageUrl":
            instance.homePageUrl = child_node.text
        elif child_node.tag == "statusPageUrl":
            instance.statusPageUrl = child_node.text
        elif child_node.tag == "healthCheckUrl":
            instance.healthCheckUrl = child_node.text
        elif child_node.tag == "secureHealthCheckUrl":
            instance.secureHealthCheckUrl = child_node.text
        elif child_node.tag == "vipAddress":
            instance.vipAddress = child_node.text
        elif child_node.tag == "secureVipAddress":
            instance.secureVipAddress = child_node.text
        elif child_node.tag == "countryId":
            instance.countryId = int(child_node.text)
        elif child_node.tag == "dataCenterInfo":
            instance.dataCenterInfo = DataCenterInfo(name=child_node.text, className=child_node.attrib["class"])
        elif child_node.tag == "hostName":
            instance.hostName = child_node.text
        elif child_node.tag == "status":
            instance.status = child_node.text
        elif child_node.tag == "overriddenstatus":
            instance.overriddenstatus = child_node.text
        elif child_node.tag == "leaseInfo":
            instance.leaseInfo = _build_lease_info(child_node)
        elif child_node.tag == "isCoordinatingDiscoveryServer":
            instance.isCoordinatingDiscoveryServer = (child_node.text == "true")
        elif child_node.tag == "metadata":
            instance.metadata = _build_metadata(child_node)
        elif child_node.tag == "lastUpdatedTimestamp":
            instance.lastUpdatedTimestamp = long(child_node.text)
        elif child_node.tag == "lastDirtyTimestamp":
            instance.lastDirtyTimestamp = long(child_node.text)
        elif child_node.tag == "actionType":
            instance.actionType = child_node.text
        elif child_node.tag == "asgName":
            instance.asgName = child_node.text

    return instance


def _build_metadata(xml_node):
    metadata = {}
    for child_node in list(xml_node):
        metadata[child_node.tag] = child_node.text
    return metadata


def _build_lease_info(xml_node):
    leaseInfo = LeaseInfo()
    for child_node in list(xml_node):
        if child_node.tag == "renewalIntervalInSecs":
            leaseInfo.renewalIntervalInSecs = int(child_node.text)
        elif child_node.tag == "durationInSecs":
            leaseInfo.durationInSecs = int(child_node.text)
        elif child_node.tag == "registrationTimestamp":
            leaseInfo.registrationTimestamp = long(child_node.text)
        elif child_node.tag == "lastRenewalTimestamp":
            leaseInfo.lastRenewalTimestamp = long(child_node.text)
        elif child_node.tag == "renewalTimestamp":
            leaseInfo.renewalTimestamp = long(child_node.text)
        elif child_node.tag == "evictionTimestamp":
            leaseInfo.evictionTimestamp = long(child_node.text)
        elif child_node.tag == "serviceUpTimestamp":
            leaseInfo.serviceUpTimestamp = long(child_node.text)

    return leaseInfo


def _build_port(xml_node):
    port = PortWrapper()
    port.port = int(xml_node.text)
    port.enabled = (xml_node.attrib["enabled"] == "true")
    return port


def get_delta(eureka_server, regions=[]):
    return _get_applications_(_format_url(eureka_server) + "apps/delta", regions)


def get_vip(eureka_server, vip, regions=[]):
    return _get_applications_(_format_url(eureka_server) + "vips/" + vip, regions)


def get_secure_vip(eureka_server, svip, regions=[]):
    return _get_applications_(_format_url(eureka_server) + "svips/" + svip, regions)


def get_application(eureka_server, app_name):
    url = _format_url(eureka_server) + "apps/" + app_name
    txt = http_client.load(url, timeout=_DEFAULT_TIME_OUT)
    return _build_application(ElementTree.fromstring(txt))


def get_app_instance(eureka_server, app_name, instance_id):
    return _get_instance_(_format_url(eureka_server) + "apps/%s/%s" % (app_name, instance_id))


def get_instance(eureka_server, instance_id):
    return _get_instance_(_format_url(eureka_server) + "instances/" + instance_id)


def _get_instance_(url):
    txt = http_client.load(url, timeout=_DEFAULT_TIME_OUT)
    return _build_instance(ElementTree.fromstring(txt))


def _current_time_millis():
    return int(time.time() * 1000)


"""====================== Registry Client ======================================="""


class RegistryClient:
    """Eureka client for spring cloud"""

    def __init__(self,
                 eureka_server=_DEFAULT_EUREKA_SERVER_URL,
                 app_name="",
                 instance_id="",
                 instance_host="",
                 instance_ip="",
                 instance_port=_DEFAULT_INSTNACE_PORT,
                 instance_unsecure_port_enabled=True,
                 instance_secure_port=_DEFAULT_INSTNACE_SECURE_PORT,
                 instance_secure_port_enabled=False,
                 countryId=1,  # @deprecaded
                 data_center_name=_DEFAULT_DATA_CENTER_INFO,  # Netflix, Amazon, MyOwn
                 renewal_interval_in_secs=_RENEWAL_INTERVAL_IN_SECS,
                 duration_in_secs=_DURATION_IN_SECS,
                 home_page_url="",
                 status_page_url="",
                 health_check_url="",
                 secure_health_check_url="",
                 vip_adr="",
                 secure_vip_addr="",
                 is_coordinating_discovery_server=False,
                 metadata={}):
        assert eureka_server is not None and eureka_server != "", "eureka server must be specified."
        assert app_name is not None and app_name != "", "application name must be specified."
        assert instance_port > 0, "port is unvalid"
        assert isinstance(metadata, dict), "metadata must be dict"

        self.__net_lock = RLock()
        self.__eureka_servers = eureka_server.split(",")

        def try_to_get_client_ip(url):
            if instance_host == "" and instance_ip == "":
                self.__instance_host = self.__instance_ip = RegistryClient.__get_instance_ip(url)
            elif instance_host != "" and instance_ip == "":
                self.__instance_host = instance_host
                if RegistryClient.__is_ip(instance_host):
                    self.__instance_ip = instance_host
                else:
                    self.__instance_ip = RegistryClient.__get_instance_ip(url)
            else:
                self.__instance_host = instance_ip
                self.__instance_ip = instance_ip

        self.__try_all_eureka_server(try_to_get_client_ip)

        mdata = {
            'management.port': str(instance_port)
        }
        mdata.update(metadata)
        self.__instance = {
            'instanceId': instance_id if instance_id != "" else "%s:%s:%d" % (self.__instance_host, app_name.lower(), instance_port),
            'hostName': self.__instance_host,
            'app': app_name.upper(),
            'ipAddr': self.__instance_ip,
            'port': {
                '$': instance_port,
                '@enabled': str(instance_unsecure_port_enabled).lower()
            },
            'securePort': {
                '$': instance_secure_port,
                '@enabled': str(instance_secure_port_enabled).lower()
            },
            'countryId': countryId,
            'dataCenterInfo': {
                '@class': _DEFAULT_DATA_CENTER_INFO_CLASS,
                'name': data_center_name
            },
            'leaseInfo': {
                'renewalIntervalInSecs': renewal_interval_in_secs,
                'durationInSecs': duration_in_secs,
                'registrationTimestamp': 0,
                'lastRenewalTimestamp': 0,
                'evictionTimestamp': 0,
                'serviceUpTimestamp': 0
            },
            'metadata': mdata,
            'homePageUrl': RegistryClient.__format_url(home_page_url, self.__instance_host, instance_port),
            'statusPageUrl': RegistryClient.__format_url(status_page_url, self.__instance_host, instance_port, "info"),
            'healthCheckUrl': RegistryClient.__format_url(health_check_url, self.__instance_host, instance_port, "health"),
            'secureHealthCheckUrl': secure_health_check_url,
            'vipAddress': vip_adr if vip_adr != "" else app_name.lower(),
            'secureVipAddress': secure_vip_addr if secure_vip_addr != "" else app_name.lower(),
            'isCoordinatingDiscoveryServer': str(is_coordinating_discovery_server).lower()
        }

        self.__alive = False
        self.__heart_beat_timer = Timer(renewal_interval_in_secs, self.__heart_beat)
        self.__heart_beat_timer.daemon = True

    def __try_all_eureka_server(self, fun):
        with self.__net_lock:
            untry_servers = self.__eureka_servers
            tried_servers = []
            ok = False
            while len(untry_servers) > 0:
                url = untry_servers[0].strip()
                try:
                    fun(url)
                except (http_client.HTTPError, http_client.URLError):
                    _logger.warn("Eureka server [%s] is down, use next url to try." % url)
                    tried_servers.append(url)
                    untry_servers = untry_servers[1:]
                else:
                    ok = True
                    break
            if len(tried_servers) > 0:
                untry_servers.extend(tried_servers)
                self.__eureka_servers = untry_servers
            if not ok:
                raise http_client.URLError("All eureka servers are down!")

    @staticmethod
    def __format_url(url, host, port, defalut_ctx=""):
        if url != "":
            if url.startswith('http'):
                _url = url
            elif url.startswith('/'):
                _url = 'http://%s:%d%s' % (host, port, url)
            else:
                _url = 'http://%s:%d/%s' % (host, port, url)
        else:
            _url = 'http://%s:%d/%s' % (host, port, defalut_ctx)
        return _url

    @staticmethod
    def __is_ip(ip_str):
        return re.match(r'^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$', ip_str)

    @staticmethod
    def __get_instance_ip(eureka_server):
        url_obj = http_client.parse_url(eureka_server)
        target_ip = url_obj["host"]
        target_port = url_obj["port"]
        if target_port is None:
            if url_obj["schema"] == "http":
                target_port = 80
            else:
                target_port = 443

        if url_obj["ipv6"] is not None:
            target_ip = url_obj["ipv6"]
            socket_family = socket.AF_INET6
        else:
            socket_family = socket.AF_INET

        s = socket.socket(socket_family, socket.SOCK_DGRAM)
        s.connect((target_ip, target_port))
        ip = s.getsockname()[0]
        s.close()
        return ip

    def register(self, status=INSTANCE_STATUS_UP, overriddenstatus=INSTANCE_STATUS_UNKNOWN):
        self.__instance["status"] = status
        self.__instance["overriddenstatus"] = overriddenstatus
        self.__instance["lastUpdatedTimestamp"] = str(_current_time_millis())
        self.__instance["lastDirtyTimestamp"] = str(_current_time_millis())
        try:
            self.__try_all_eureka_server(lambda url: _register(url, self.__instance))
        except:
            _logger.exception("error!")
        else:
            self.__alive = True

    def cancel(self):
        try:
            self.__try_all_eureka_server(lambda url: cancel(url, self.__instance["app"], self.__instance["instanceId"]))
        except:
            _logger.exception("error!")
        else:
            self.__alive = False

    def send_heart_beat(self, overridden_status=""):
        try:
            self.__try_all_eureka_server(lambda url: send_heart_beat(url, self.__instance["app"],
                                                                     self.__instance["instanceId"], self.__instance["lastDirtyTimestamp"],
                                                                     status=self.__instance["status"], overriddenstatus=overridden_status))
        except:
            _logger.exception("Error!")
            _logger.info("Cannot send heartbeat to server, try to register")
            self.register()

    def status_update(self, new_status):
        self.__instance["status"] = new_status
        try:
            self.__try_all_eureka_server(lambda url: status_update(url, self.__instance["app"], self.__instance["instanceId"],
                                                                   self.__instance["lastDirtyTimestamp"], new_status))
        except:
            _logger.exception("error!")

    def delete_status_override(self):
        self.__try_all_eureka_server(lambda url: delete_status_override(
            url, self.__instance["app"], self.__instance["instanceId"], self.__instance["lastDirtyTimestamp"]))

    def start(self):
        _logger.debug("start to registry client...")
        self.register()
        self.__heart_beat_timer.start()

    def stop(self):
        if self.__alive:
            _logger.debug("stopping client...")
            if self.__heart_beat_timer.isAlive():
                self.__heart_beat_timer.cancel()
            self.register(status=INSTANCE_STATUS_DOWN)
            self.cancel()

    def __heart_beat(self):
        while True:
            _logger.debug("sending heart beat to spring cloud server ")
            self.send_heart_beat()
            time.sleep(self.__instance["leaseInfo"]["renewalIntervalInSecs"])


__cache_registry_clients = {}
__cache_registry_clients_lock = RLock()


def init_registry_client(eureka_server=_DEFAULT_EUREKA_SERVER_URL,
                         app_name="",
                         instance_id="",
                         instance_host="",
                         instance_ip="",
                         instance_port=_DEFAULT_INSTNACE_PORT,
                         instance_unsecure_port_enabled=True,
                         instance_secure_port=_DEFAULT_INSTNACE_SECURE_PORT,
                         instance_secure_port_enabled=False,
                         countryId=1,  # @deprecaded
                         data_center_name=_DEFAULT_DATA_CENTER_INFO,  # Netflix, Amazon, MyOwn
                         renewal_interval_in_secs=_RENEWAL_INTERVAL_IN_SECS,
                         duration_in_secs=_DURATION_IN_SECS,
                         home_page_url="",
                         status_page_url="",
                         health_check_url="",
                         secure_health_check_url="",
                         vip_adr="",
                         secure_vip_addr="",
                         is_coordinating_discovery_server=False,
                         metadata={}):
    with __cache_registry_clients_lock:
        client = RegistryClient(eureka_server=eureka_server,
                                app_name=app_name,
                                instance_id=instance_id,
                                instance_host=instance_host,
                                instance_ip=instance_ip,
                                instance_port=instance_port,
                                instance_unsecure_port_enabled=instance_unsecure_port_enabled,
                                instance_secure_port=instance_secure_port,
                                instance_secure_port_enabled=instance_secure_port_enabled,
                                countryId=countryId,
                                data_center_name=data_center_name,
                                renewal_interval_in_secs=renewal_interval_in_secs,
                                duration_in_secs=duration_in_secs,
                                home_page_url=home_page_url,
                                status_page_url=status_page_url,
                                health_check_url=health_check_url,
                                secure_health_check_url=secure_health_check_url,
                                vip_adr=vip_adr,
                                secure_vip_addr=secure_vip_addr,
                                is_coordinating_discovery_server=is_coordinating_discovery_server,
                                metadata=metadata)
        __cache_registry_clients[os.getpid()] = client
        client.start()
        return client


def get_registry_client():
    # type () -> RegistryClient
    __cache_key = os.getpid()
    with __cache_registry_clients_lock:
        if __cache_key in __cache_registry_clients:
            return __cache_registry_clients[__cache_key]
        else:
            return None


"""======================== Cached Discovery Client ============================"""


class DiscoveryClient:
    """Discover the apps registered in spring cloud server, this class will do some cached, if you want to get the apps immediatly, use the global functions"""

    def __init__(self, eureka_server, regions=None, renewal_interval_in_secs=_RENEWAL_INTERVAL_IN_SECS, ha_strategy=HA_STRATEGY_RANDOM):
        assert ha_strategy in [HA_STRATEGY_RANDOM, HA_STRATEGY_STICK, HA_STRATEGY_OTHER], "do not support strategy %d " % ha_strategy
        self.__eureka_servers = eureka_server.split(",")
        self.__regions = regions if regions is not None else []
        self.__cache_time_in_secs = renewal_interval_in_secs
        self.__applications = None
        self.__delta = None
        self.__ha_strategy = ha_strategy
        self.__ha_cache = {}
        self.__timer = Timer(self.__cache_time_in_secs, self.__heartbeat)
        self.__timer.daemon = True
        self.__application_mth_lock = RLock()
        self.__net_lock = RLock()

    def __heartbeat(self):
        while True:
            self.__fetch_delta()
            time.sleep(self.__cache_time_in_secs)

    @property
    def applications(self):
        with self.__application_mth_lock:
            if self.__applications is None:
                self.__pull_full_registry()
            return self.__applications

    def __try_all_eureka_server(self, fun):
        with self.__net_lock:
            untry_servers = self.__eureka_servers
            tried_servers = []
            ok = False
            while len(untry_servers) > 0:
                url = untry_servers[0].strip()
                try:
                    fun(url)
                except (http_client.HTTPError, http_client.URLError):
                    _logger.warn("Eureka server [%s] is down, use next url to try." % url)
                    tried_servers.append(url)
                    untry_servers = untry_servers[1:]
                else:
                    ok = True
                    break
            if len(tried_servers) > 0:
                untry_servers.extend(tried_servers)
                self.__eureka_servers = untry_servers
            if not ok:
                raise http_client.URLError("All eureka servers are down!")

    def __pull_full_registry(self):
        def do_pull(url):  # the actual function body
            self.__applications = get_applications(url, self.__regions)
            self.__delta = self.__applications
        self.__try_all_eureka_server(do_pull)

    def __fetch_delta(self):
        def do_fetch(url):
            if self.__applications is None or len(self.__applications.applications) == 0:
                self.__pull_full_registry()
                return
            delta = get_delta(url, self.__regions)
            _logger.debug("delta got: v.%s::%s" % (delta.versionsDelta, delta.appsHashcode))
            if self.__delta is not None \
                    and delta.versionsDelta == self.__delta.versionsDelta \
                    and delta.appsHashcode == self.__delta.appsHashcode:
                return
            self.__merge_delta(delta)
            self.__delta = delta
            if not self.__is_hash_match():
                self.__pull_full_registry()
        self.__try_all_eureka_server(do_fetch)

    def __is_hash_match(self):
        app_hash = self.__get_applications_hash()
        _logger.debug("check hash, local[%s], remote[%s]" % (app_hash, self.__delta.appsHashcode))
        return app_hash == self.__delta.appsHashcode

    def __merge_delta(self, delta):
        _logger.debug("merge delta...length of application got from delta::%d" % len(delta.applications))
        for application in delta.applications:
            for instance in application.instances:
                _logger.debug("instance [%s] has %s" % (instance.instanceId, instance.actionType))
                if instance.actionType in (ACTION_TYPE_ADDED, ACTION_TYPE_MODIFIED):
                    existingApp = self.applications.get_application(application.name)
                    if existingApp is None:
                        self.applications.add_application(application)
                    else:
                        existingApp.update_instance(instance)
                elif instance.actionType == ACTION_TYPE_DELETED:
                    existingApp = self.applications.get_application(application.name)
                    if existingApp is None:
                        self.applications.add_application(application)
                    existingApp.remove_instance(instance)

    def __get_applications_hash(self):
        app_hash = ""
        app_status_count = {}
        for application in self.__applications.applications:
            for instance in application.instances:
                if instance.status not in app_status_count:
                    app_status_count[instance.status.upper()] = 0
                app_status_count[instance.status.upper()] = app_status_count[instance.status.upper()] + 1

        sorted_app_status_count = sorted(app_status_count.items(), key=lambda item: item[0])
        for item in sorted_app_status_count:
            app_hash = app_hash + "%s_%d_" % (item[0], item[1])
        return app_hash

    def walk_nodes_async(self, app_name="", service="", prefer_ip=False, prefer_https=False, walker=None, on_success=None, on_error=None):
        def async_thread_target():
            try:
                res = self.walk_nodes(app_name=app_name, service=service, prefer_ip=prefer_ip, prefer_https=prefer_https, walker=walker)
                if on_success is not None and (inspect.isfunction(on_success) or inspect.ismethod(on_success)):
                    on_success(res)
            except http_client.HTTPError as e:
                if on_error is not None and (inspect.isfunction(on_error) or inspect.ismethod(on_error)):
                    on_error(e)

        async_thread = Thread(target=async_thread_target)
        async_thread.daemon = True
        async_thread.start()

    def walk_nodes(self, app_name="", service="", prefer_ip=False, prefer_https=False, walker=None):
        assert app_name is not None and app_name != "", "application_name should not be null"
        assert inspect.isfunction(walker) or inspect.ismethod(walker), "walker must be a method or function"
        error_nodes = []
        app_name = app_name.upper()
        node = self.__get_available_service(app_name)

        while node is not None:
            try:
                url = self.__generate_service_url(node, prefer_ip, prefer_https)
                if service.startswith("/"):
                    url = url + service[1:]
                else:
                    url = url + service
                _logger.debug("service url::" + url)
                return walker(url)
            except (http_client.HTTPError, http_client.URLError):
                _logger.warn("do service %s in node [%s] error, use next node." % (service, node.instanceId))
                error_nodes.append(node.instanceId)
                node = self.__get_available_service(app_name, error_nodes)

        raise http_client.URLError("Try all up instances in registry, but all fail")

    def do_service_async(self, app_name="", service="", return_type="string",
                         prefer_ip=False, prefer_https=False,
                         on_success=None, on_error=None,
                         method="GET", headers=None,
                         data=None, timeout=_DEFAULT_TIME_OUT,
                         cafile=None, capath=None, cadefault=False, context=None):
        def async_thread_target():
            try:
                res = self.do_service(app_name=app_name,
                                      service=service, return_type=return_type,
                                      prefer_ip=prefer_ip, prefer_https=prefer_https,
                                      method=method, headers=headers,
                                      data=data, timeout=timeout,
                                      cafile=cafile, capath=capath,
                                      cadefault=cadefault, context=context)
                if on_success is not None and (inspect.isfunction(on_success) or inspect.ismethod(on_success)):
                    on_success(res)
            except http_client.HTTPError as e:
                if on_error is not None and (inspect.isfunction(on_error) or inspect.ismethod(on_error)):
                    on_error(e)

        async_thread = Thread(target=async_thread_target)
        async_thread.daemon = True
        async_thread.start()

    def do_service(self, app_name="", service="", return_type="string",
                   prefer_ip=False, prefer_https=False,
                   method="GET", headers=None,
                   data=None, timeout=_DEFAULT_TIME_OUT,
                   cafile=None, capath=None, cadefault=False, context=None):
        def walk_using_urllib(url):
            req = http_client.Request(url)
            req.get_method = lambda: method
            heads = headers if headers is not None else {}
            for k, v in heads.items():
                req.add_header(k, v)

            res_txt = http_client.load(req, data=data, timeout=timeout, cafile=cafile, capath=capath, cadefault=cadefault, context=context)
            if return_type.lower() in ("json", "dict", "dictionary"):
                return json.loads(res_txt)
            else:
                return res_txt
        return self.walk_nodes(app_name, service, prefer_ip, prefer_https, walk_using_urllib)

    def __get_available_service(self, application_name, ignore_instance_ids=None):
        app = self.applications.get_application(application_name)
        if app is None:
            return None
        up_instances = []
        if ignore_instance_ids is None or len(ignore_instance_ids) == 0:
            up_instances.extend(app.up_instances)
        else:
            for ins in app.up_instances:
                if ins.instanceId not in ignore_instance_ids:
                    up_instances.append(ins)

        if len(up_instances) == 0:
            # no up instances
            return None
        elif len(up_instances) == 1:
            # only one available instance, then doesn't matter which strategy is.
            instance = up_instances[0]
            self.__ha_cache[application_name] = instance.instanceId
            return instance

        def random_one(instances):
            if len(instances) == 1:
                idx = 0
            else:
                idx = random.randint(0, len(instances) - 1)
            selected_instance = instances[idx]
            self.__ha_cache[application_name] = selected_instance.instanceId
            return selected_instance

        if self.__ha_strategy == HA_STRATEGY_RANDOM:
            return random_one(up_instances)
        elif self.__ha_strategy == HA_STRATEGY_STICK:
            if application_name in self.__ha_cache:
                cache_id = self.__ha_cache[application_name]
                cahce_instance = app.get_instance(cache_id)
                if cahce_instance is not None and cahce_instance.status == INSTANCE_STATUS_UP:
                    return cahce_instance
                else:
                    return random_one(up_instances)
            else:
                return random_one(up_instances)
        elif self.__ha_strategy == HA_STRATEGY_OTHER:
            if application_name in self.__ha_cache:
                cache_id = self.__ha_cache[application_name]
                other_instances = []
                for up_instance in up_instances:
                    if up_instance.instanceId != cache_id:
                        other_instances.append(up_instance)
                return random_one(other_instances)
            else:
                return random_one(up_instances)
        else:
            return None

    def __generate_service_url(self, instance, prefer_ip, prefer_https):
        if instance is None:
            return None
        schema = "http"
        port = 0
        if instance.port.port and not instance.securePort.enabled:
            schema = "http"
            port = instance.port.port
        elif not instance.port.port and instance.securePort.enabled:
            schema = "https"
            port = instance.securePort.port
        elif instance.port.port and instance.securePort.enabled:
            if prefer_https:
                schema = "https"
                port = instance.securePort.port
            else:
                schema = "http"
                port = instance.port.port
        else:
            assert False, "generate_service_url error: No port is available"

        host = instance.ipAddr if prefer_ip else instance.hostName

        return "%s://%s:%d/" % (schema, host, port)

    def start(self):
        self.__pull_full_registry()
        self.__timer.start()

    def stop(self):
        if self.__timer.isAlive():
            self.__timer.cancel()


__cache_discovery_clients = {}
__cache_discovery_clients_lock = RLock()


def init_discovery_client(eureka_server=_DEFAULT_EUREKA_SERVER_URL, regions=[], renewal_interval_in_secs=_RENEWAL_INTERVAL_IN_SECS, ha_strategy=HA_STRATEGY_RANDOM):
    __cache_key = os.getpid()
    with __cache_discovery_clients_lock:
        assert __cache_key not in __cache_discovery_clients, "Client has already been initialized."
        cli = DiscoveryClient(eureka_server, regions=regions, renewal_interval_in_secs=renewal_interval_in_secs, ha_strategy=ha_strategy)
        cli.start()
        __cache_discovery_clients[__cache_key] = cli
        return cli


def get_discovery_client():
    # type: (str) -> DiscoveryClient
    __cache_key = os.getpid()
    with __cache_discovery_clients_lock:
        if __cache_key in __cache_discovery_clients:
            return __cache_discovery_clients[__cache_key]
        else:
            return None


def init(eureka_server=_DEFAULT_EUREKA_SERVER_URL,
         regions=[],
         app_name="",
         instance_id="",
         instance_host="",
         instance_ip="",
         instance_port=_DEFAULT_INSTNACE_PORT,
         instance_unsecure_port_enabled=True,
         instance_secure_port=_DEFAULT_INSTNACE_SECURE_PORT,
         instance_secure_port_enabled=False,
         countryId=1,  # @deprecaded
         data_center_name=_DEFAULT_DATA_CENTER_INFO,  # Netflix, Amazon, MyOwn
         renewal_interval_in_secs=_RENEWAL_INTERVAL_IN_SECS,
         duration_in_secs=_DURATION_IN_SECS,
         home_page_url="",
         status_page_url="",
         health_check_url="",
         secure_health_check_url="",
         vip_adr="",
         secure_vip_addr="",
         is_coordinating_discovery_server=False,
         metadata={},
         ha_strategy=HA_STRATEGY_RANDOM):
    registry_client = init_registry_client(eureka_server=eureka_server,
                                           app_name=app_name,
                                           instance_id=instance_id,
                                           instance_host=instance_host,
                                           instance_ip=instance_ip,
                                           instance_port=instance_port,
                                           instance_unsecure_port_enabled=instance_unsecure_port_enabled,
                                           instance_secure_port=instance_secure_port,
                                           instance_secure_port_enabled=instance_secure_port_enabled,
                                           countryId=countryId,
                                           data_center_name=data_center_name,
                                           renewal_interval_in_secs=renewal_interval_in_secs,
                                           duration_in_secs=duration_in_secs,
                                           home_page_url=home_page_url,
                                           status_page_url=status_page_url,
                                           health_check_url=health_check_url,
                                           secure_health_check_url=secure_health_check_url,
                                           vip_adr=vip_adr,
                                           secure_vip_addr=secure_vip_addr,
                                           is_coordinating_discovery_server=is_coordinating_discovery_server,
                                           metadata=metadata)
    discovery_client = init_discovery_client(eureka_server,
                                             regions=regions,
                                             renewal_interval_in_secs=renewal_interval_in_secs,
                                             ha_strategy=ha_strategy)
    return registry_client, discovery_client


def walk_nodes_async(app_name="", service="", prefer_ip=False, prefer_https=False, walker=None, on_success=None, on_error=None):
    cli = get_discovery_client()
    if cli is None:
        raise Exception("Discovery Client has not initialized. ")
    cli.walk_nodes_async(app_name=app_name, service=service,
                         prefer_ip=prefer_ip, prefer_https=prefer_https,
                         walker=walker, on_success=on_success, on_error=on_error)


def walk_nodes(app_name="", service="", prefer_ip=False, prefer_https=False, walker=None):
    cli = get_discovery_client()
    if cli is None:
        raise Exception("Discovery Client has not initialized. ")
    return cli.walk_nodes(app_name=app_name, service=service,
                          prefer_ip=prefer_ip, prefer_https=prefer_https, walker=walker)


def do_service_async(app_name="", service="", return_type="string",
                     prefer_ip=False, prefer_https=False,
                     on_success=None, on_error=None,
                     method="GET", headers=None,
                     data=None, timeout=_DEFAULT_TIME_OUT,
                     cafile=None, capath=None, cadefault=False, context=None):
    cli = get_discovery_client()
    if cli is None:
        raise Exception("Discovery Client has not initialized. ")
    cli.do_service_async(app_name=app_name, service=service, return_type=return_type,
                         prefer_ip=prefer_ip, prefer_https=prefer_https,
                         on_success=on_success, on_error=on_error,
                         method=method, headers=headers,
                         data=data, timeout=timeout,
                         cafile=cafile, capath=capath,
                         cadefault=cadefault, context=context)


def do_service(app_name="", service="", return_type="string",
               prefer_ip=False, prefer_https=False,
               method="GET", headers=None,
               data=None, timeout=_DEFAULT_TIME_OUT,
               cafile=None, capath=None, cadefault=False, context=None):
    cli = get_discovery_client()
    if cli is None:
        raise Exception("Discovery Client has not initialized. ")
    return cli.do_service(app_name=app_name, service=service, return_type=return_type,
                          prefer_ip=prefer_ip, prefer_https=prefer_https,
                          method=method, headers=headers,
                          data=data, timeout=timeout,
                          cafile=cafile, capath=capath,
                          cadefault=cadefault, context=context)


def stop():
    register_cli = get_registry_client()
    if register_cli is not None:
        register_cli.stop()
    discovery_client = get_discovery_client()
    if discovery_client is not None:
        discovery_client.stop()


@atexit.register
def _cleanup_before_exist():
    if len(__cache_registry_clients) > 0:
        _logger.debug("cleaning up registry clients")
        for k, cli in __cache_registry_clients.items():
            _logger.debug("try to stop cache registry client [%s] this will also unregister this client from the eureka server" % k)
            cli.stop()
    if len(__cache_discovery_clients) > 0:
        _logger.debug("cleaning up discovery clients")
        for k, cli in __cache_discovery_clients.items():
            _logger.debug("try to stop cache discovery client [%s] this will also unregister this client from the eureka server" % k)
            cli.stop()




# python实例的缓存
from api.exceptions import IllegalArgumentException
from api.handler.IHandler import IHandler
from api.handler.ce.AbstractBatchCE import AbstractBatchCE
from api.handler.crawl.AbstractBatchCrawl import AbstractBatchCrawl
from api.handler.crawl.AbstractSimpleCrawl import AbstractSimpleCrawl
from api.handler.ce.ISimpleScriptCE import ISimpleScriptCE
from api.handler.crawl.ISimpleScriptCrawl import ISimpleScriptCrawl


PYTHON_INSTANCE_CACHE = {}

def _is_subclass(v,clazz)->bool:
    cc = None
    if clazz=="ISimpleScriptCE":
        cc = ISimpleScriptCE
    elif  clazz=="ISimpleScriptCrawl":
        cc = ISimpleScriptCrawl
    elif clazz == "ISampleScriptCE":
        from api.handler.ce.ISampleScriptCE import ISampleScriptCE
        cc = ISampleScriptCE
    elif clazz == "ISampleScriptCrawl":
        from api.handler.crawl.ISampleScriptCrawl import ISampleScriptCrawl
        cc = ISampleScriptCrawl
    elif clazz == "AbstractBatchCE":
        cc = AbstractBatchCE
    elif clazz == "AbstractBatchCrawl":
        cc = AbstractBatchCrawl
    elif clazz == "AbstractSimpleCrawl":
        cc = AbstractSimpleCrawl
    elif clazz == "IHandler":
        cc = IHandler

    if cc is None:
        return False
    return issubclass(v,cc)

def _find_class(context:dict,clazz):
    script_class = None
    for k, v in context.items():
        if k != clazz and isinstance(v, type) and _is_subclass(v,clazz):
            script_class = v
    return script_class

def load_instance(name:str, code_source:str,clazz):
    if  name in PYTHON_INSTANCE_CACHE:
        # 说明在缓存中,直接返回
        return PYTHON_INSTANCE_CACHE[name]

    # 说明还没有实例化,
    if code_source is None or len(code_source)==0:
        raise IllegalArgumentException(f"加载{name} 失败,源码为空")
    else:
        # 说明是有东西的,开始尝试加载
        context = {}
        from api.model.FdInputPipe import FdInputPipe
        from api.model.BatchOutputPipe import BatchOutputPipe

        context["ISimpleScriptCrawl"] = ISimpleScriptCrawl
        context["ISimpleScriptCE"] = ISimpleScriptCE
        from api.handler.crawl.ISampleScriptCrawl import ISampleScriptCrawl
        context["ISampleScriptCrawl"] = ISampleScriptCrawl
        from api.handler.ce.ISampleScriptCE import ISampleScriptCE
        context["ISampleScriptCE"] = ISampleScriptCE
        context["AbstractBatchCE"] = AbstractBatchCE
        context["AbstractBatchCrawl"] = AbstractBatchCrawl
        context["AbstractSimpleCrawl"] = AbstractSimpleCrawl
        context["FdInputPipe"] = FdInputPipe
        context["BatchOutputPipe"] = BatchOutputPipe

        # 动态加载源码
        exec(code_source.strip(), context)  # None

        # 找到类定义
        script_class = _find_class(context,clazz)

        if  script_class is None:
            # 说明没有找到,报错
            raise IllegalArgumentException(f"加载{name}的源码失败,没有继承自ISimpleScriptCE/ISimpleScriptCrawl")

        # 放入缓存
        PYTHON_INSTANCE_CACHE[name] = script_class()

        # 返回结果
        return PYTHON_INSTANCE_CACHE[name]

def destroy_instance(name:str):
    """
    清理实例
    :param name:
    :return:
    """
    if  name in PYTHON_INSTANCE_CACHE:
        # 如果存在,就销毁掉
        tmp = PYTHON_INSTANCE_CACHE[name]
        del PYTHON_INSTANCE_CACHE[name]
        del tmp



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/22
import threading

from sqlalchemy import create_engine
from sqlalchemy.orm import Session, sessionmaker
from urllib3.util import url, Url

from api.exceptions import IllegalArgumentException
from api.model.FactDatasource import JdbcTypeEnum, JdbcDriverEnum, FactDatasourceTypeEnum
from config.annotation import Value
from factdatasource.dao.FactDatasource import DatasourceListener, MultipleBaseDatasource, BaseDatasource, \
    customer_dao_context_holder
from factdatasource.execptions import DatasourceNotFoundException, NotSupportedFDException, FDExecuteException, \
    FDConfigException
from libs import utils
from libs.utils import Singleton


class MultipleJdbcDatasource(DatasourceListener, MultipleBaseDatasource, Singleton):
    """
    管理fd中的所有JDBC类型的数据源
    """

    DEFAULT_JDBC_DATASOURCE = '_ficus4py_default_jdbc_datasource'

    def __init__(self):
        # source_name --> JdbcDatasource
        self._target_dataSources = dict()

    def get_datasource_type(self):
        """
        获取数据源类型
        :return:
        """
        return FactDatasourceTypeEnum.JDBC

    def add_datasource_type(self, source_name: str, url: str, credentials: str):
        """
        添加一个数据源
        :param source_name:
        :param url:
        :param credentials:
        :return:
        """
        if not source_name or not url:
            raise IllegalArgumentException(f'添加数据源参数错误:source_name、url都不能为空')

        for target in self._target_dataSources.values():
            if target.url == url and target.credentials == credentials:
                # url 和 credentials 都相同，说明是同一个数据库连接就不再重复创建数据源了
                target.add_source_name(source_name)
                self._target_dataSources[source_name] = target
                return

        # 开始创建数据源
        jdbc_data_source = JdbcDatasource(source_name, url, credentials)
        jdbc_data_source.start()
        self._target_dataSources[source_name] = jdbc_data_source

    def update_datasource_type(self, source_name: str, url: str, credentials: str):
        """
        修改一个数据源
        :param source_name:
        :param url:
        :param credentials:
        :return:
        """
        if not source_name or not url:
            raise IllegalArgumentException(f'修改数据源参数错误:source_name、url都不能为空')

        self.delete_datasource_type(source_name)
        self.add_datasource_type(source_name, url, credentials)

    def delete_datasource_type(self, source_name: str):
        """
        删除一个数据源
        :param source_name:
        :return:
        """
        if not source_name:
            raise IllegalArgumentException(f'删除数据源参数错误:source_name不能为空')

        target: JdbcDatasource = self._target_dataSources.pop(source_name)
        if target:
            if target.only_one_source():
                target.close_client()
            else:
                target.remove_source_name(source_name)

    def get_data_source(self):
        """
        获取数据源的基本信息
        :return:
        """
        source_name = self.determine_current_lookup_key()
        if not source_name:
            raise FDExecuteException('未设置操作源,无法获取数据源信息。')
        if source_name in self._target_dataSources.keys():
            return self._target_dataSources[source_name]
        raise DatasourceNotFoundException('未发现数据源%s,请添加后再获取。' % source_name)

    def determine_current_lookup_key(self):
        """JDBC中需要设置默认的数据源，所以这里需要重写该方法"""
        source_name = customer_dao_context_holder.get_source()
        if source_name is None:
            return self.DEFAULT_JDBC_DATASOURCE
        return source_name

    def get_client(self):
        """
        获取数据库操作的client
        :return:
        """
        target: JdbcDatasource = self.get_data_source()
        return target.get_client()

    def close_client(self):
        """
        关闭客户端，关闭连接
        :return:
        """
        target: JdbcDatasource = self.get_data_source()
        return target.close_client()


class JdbcDatasource(BaseDatasource):
    """
    JDBC数据源，存储该数据源基本信息
    """

    def __init__(self, source_name, url, credentials):
        self.init(source_name, url, credentials)
        self.__lock = threading.Lock()
        self.__client: Session = None
        self.__engine = None

    def start(self):
        """
        分析url并开始连接数据库
        :return:
        """
        sqlalchemy_url = self._parse_url()
        self.__engine = create_engine(sqlalchemy_url, case_sensitive=False, echo=True, encoding='utf-8')

    def get_client(self):
        if not self.__client:
            with self.__lock:
                if not self.__client:
                    jdbc_session = sessionmaker(bind=self.__engine, autocommit=True)
                    self.__client = jdbc_session()
        return self.__client

    def close_client(self):
        if self.__client:
            self.__client.close()

    def _parse_url(self):
        """
        分析url，并构造成能直接连接sqlalchemy的url
        :return:
        """
        if not self._url:
            return ''
        jdbc_type = utils.get_jdbc_type(self._url)
        if not jdbc_type:
            raise NotSupportedFDException(f'不支持该连接{self._url}')

        result_url = self._structure_url(jdbc_type, self.url)
        return result_url

    def _structure_url(self, jdbc_type: JdbcDriverEnum, jdbc_url: str):
        """
        由于url都是JDBC版本的url，所以需要根据不同的版本构造我们需要的url
        :param jdbc_type:
        :param jdbc_url:
        :return:  dialect[+driver]://user:password@host/dbname[?key=value..]
        """
        if not jdbc_type or not jdbc_url:
             raise IllegalArgumentException('构造url参数错误。')

        # 支持sqlalchemy进行连接的url
        if jdbc_type == JdbcTypeEnum.MYSQL:
            sqlalchemy_scheme = JdbcDriverEnum.MYSQL_DRIVER
        elif jdbc_type == JdbcTypeEnum.ORACLE:
            sqlalchemy_scheme = JdbcDriverEnum.ORACLE_DRIVER
        elif jdbc_type == JdbcTypeEnum.POSTGRESQL:
            sqlalchemy_scheme = JdbcDriverEnum.POSTGRESQL_DRIVER
        elif jdbc_type == JdbcTypeEnum.SQL_SERVER:
            sqlalchemy_scheme = JdbcDriverEnum.SQL_SERVER_DRIVER
        elif jdbc_type == JdbcTypeEnum.VERNOX:
            sqlalchemy_scheme = JdbcDriverEnum.VERNOX_DRIVER
        else:
            raise NotSupportedFDException(f'暂不支持该数据库{jdbc_type}')

        result_url = None
        if jdbc_type == JdbcTypeEnum.SQL_SERVER:
            # JDBC的通用形式是 jdbc:sqlserver://[serverName[\instanceName][:portNumber]][;property=value[;property=value]]
            url_param = {}
            if ';' in jdbc_url:
                url_list = jdbc_url.split(';', 1)
                url_prefix = url_list[0]
                url_param_tmp = url_list[1]
                # 获取配置的参数
                for param in url_param_tmp.split(';'):
                    if '=' in param:
                        p = param.split('=', 1)
                        url_param[p[0]] = p[1]
            else:
                url_prefix = jdbc_url

            r = url.parse_url(url_prefix)
            if 'databaseName' in url_param:
                database = url_param.pop('databaseName')
            else:
                raise FDConfigException(f'数据库连接{jdbc_url}未配置databaseName值。')

            # 添加一个驱动参数
            if 'driver' not in url_param:
                url_param['driver'] = 'SQL Server'

            query = ''
            if url_param:
                for key, value in url_param.items():
                    query = query + str(key) + '=' + str(value) + '&'
                # 删除最后一个&
                query = query.rstrip('&')

            path = '/' + str(database)

            result_url = Url(scheme=sqlalchemy_scheme.value, auth=self._credentials, host=r.host, port=r.port, path=path, query=query, fragment=r.fragment)
        elif jdbc_type == JdbcTypeEnum.ORACLE:
            # JDBC的常用形式是
            # 1.普通SID方式
            # jdbc:oracle:thin:username/password@x.x.x.1:1521:SID
            # 2.普通ServiceName 方式
            # jdbc:oracle:thin:username/password@//x.x.x.1:1522/ABCD
            # 3.RAC方式  暂时不支持
            # jdbc:oracle:thin:@(DESCRIPTION=(ADDRESS_LIST=(ADDRESS=(PROTOCOL=TCP)(HOST=x.x.x.1)(PORT=1521))(ADDRESS=(PROTOCOL=TCP)(HOST=x.x.x.2)(PORT=1521)))(LOAD_BALANCE=yes)(CONNECT_DATA=(SERVER=DEDICATED)(SERVICE_NAME=xxrac)))
            base_message = None
            if '@' in jdbc_url:
                base_message = jdbc_url.split('@', 1)[1]
            else:
                base_message = jdbc_url

            host_port = None
            service_name = None
            sid = None

            if jdbc_url.startswith('//'):
                # ServiceName 方式
                if '/' in base_message:
                    host_port = base_message.split('/', 1)[0]
                    service_name = base_message.split('/', 1)[1]
                else:
                    raise FDConfigException(f'数据库连接{jdbc_url}未配置service_name值,无法解析。')
            else:
                # sid方式
                if ":" in base_message:
                    sid_message = base_message.split(':', 2)
                    if len(sid_message) == 3:
                        host_port = str(sid_message[0]) + ':' + str(sid_message[1])
                        sid = str(sid_message[2])
                    elif len(sid_message) == 2:
                        host_port = str(sid_message[0])
                        sid = str(sid_message[1])
                    else:
                        raise FDConfigException(f'数据库连接{jdbc_url}无法进行有效解析。')
                else:
                    raise FDConfigException(f'数据库连接{jdbc_url}无法进行有效解析。')

            if ':' in host_port:
                host = str(host_port.split(':', 1)[0])
                port = str(host_port.split(':', 1)[1])
            else:
                host = str(host_port)
                port = '1521'

            import cx_Oracle
            if service_name:
                dsn = cx_Oracle.makedsn(host, port, service_name=service_name)
            else:
                dsn = cx_Oracle.makedsn(host, port, sid=sid)

            credentials = str(self._credentials)+'@' if self._credentials else ''
            return f'{sqlalchemy_scheme.value}://{credentials}{dsn}'
        else:
            # 默认可以使用的比较通用的构造方式
            r = url.parse_url(jdbc_url.split(":",1)[1])
            # sqlalchemy mysql 这里中文一直是乱码，暂时先手动设置下编码
            query = r.query
            if jdbc_type == JdbcTypeEnum.MYSQL or jdbc_type == JdbcTypeEnum.VERNOX:
                if not query:
                    query = 'charset=utf8mb4'
                elif str(query).lower().find('charset=') == -1:
                    query = query + '&charset=utf8'

            result_url = Url(scheme=sqlalchemy_scheme.value, auth=self._credentials, host=r.host, port=r.port, path=r.path, query=query, fragment=r.fragment)
        return result_url.url


@Value("${jdbc.ip}")
def jdbc_ip():
    pass


@Value("${jdbc.url}")
def jdbc_url():
    pass


@Value("${jdbc.username}")
def jdbc_username():
    pass


@Value("${jdbc.password}")
def jdbc_password():
    pass


def register_jdbc_default_datasource():
    """
    注册默认的数据源
    :return:
    """
    credentials = f'{jdbc_username()}:{jdbc_password()}'
    url = jdbc_url()
    datasource = MultipleJdbcDatasource.instance()
    datasource.add_datasource_type(datasource.DEFAULT_JDBC_DATASOURCE, url, credentials)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/30
from unittest import TestCase

from factdatasource.FactDatasourceContextHolder import FactDatasourceContextHolder
from factdatasource.dao.jdbc.MultipleJdbcDatasource import MultipleJdbcDatasource


def register_default_datasource():
    # ficus数据库配置
    credentials = 'sobeyhive:$0bEyHive&2o1Six'
    url = 'jdbc:mysql://localhost:3306/ficus'
    datasource = MultipleJdbcDatasource.instance()
    datasource.add_datasource_type(datasource.DEFAULT_DATASOURCE, url, credentials)


class TestDistributedFactDatasourceProxy(TestCase):

    def test_size(self):
        register_default_datasource()
        fd_context_holder = FactDatasourceContextHolder.instance()
        fd_context1 = fd_context_holder.get_fact_datasource('test_source_1')
        size = fd_context1.size()
        print(f'数据源test_source_1长度{size}')

        fd_context2 = fd_context_holder.get_fact_datasource('test_source_2')
        size = fd_context2.size()
        print(f'数据源test_source_2长度{size}')

        print(fd_context2.collect(5))

        print(fd_context1.collect(5))



"""
Crawl的操作模块
"""
import requests
from munch import Munch

from api.exceptions import ServiceInnerException, AuthException
from client import check_instance_avaliable, do_service


def get(site, project_code, code):
    """
    获取CE
    :param site: 站点
    :param project_code: 项目名
    :param code: CE的唯一code
    :return: 如果这个ce存在,返回ce对象.否则返回空
    """
    check_instance_avaliable()

    try:
        r = do_service(f"remote/dc-service/{site}/{project_code}/{code}")
        if r is not None:
            return Munch(r)
        else:
            return None
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code<500 and e.response.status_code!=404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e

def exists(site, project_code, code):
    """
    校验CE是否存在
    :param site: 站点
    :param project_code: 项目名
    :param code: CE的唯一code
    :return: 如果这个ce存在,返回true.否则,返回false
    """
    check_instance_avaliable()

    try:
        r = do_service(f"remote/dc-service/exists/{site}/{project_code}/{code}")
        if r is not None:
            return r
        else:
            return False
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code<500 and e.response.status_code!=404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def query(site, project_code=None, code=None, crawl_type=None, page=1, size=20):
    """
    查询Crawl
    :param site: 站点
    :param project_code: 项目名
    :param code: CE的唯一code
    :param crawl_type: crawl的类型, JDBC/FILE/WEB/CUSTOM
    :param page: 分页页码,从1开始
    :param size: 分页大小,默认20
    :return:
    """
    check_instance_avaliable()

    try:
        r = do_service(f"remote/dc-service/{site}",params={'projectcode': project_code, 'code': code, 'type': crawl_type, 'page': page,
                                 'size': size})
        if r is not None:
            return Munch(r)
        else:
            return None
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code<500 and e.response.status_code!=404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def add_crawl(data_crawl):
    """
    新增一个Crawl
    :param data_crawl: crawl对象
    :return: 如果添加失败,就抛出异常,否则无返回
    """
    if data_crawl is None:
        return

    check_instance_avaliable()

    request = None
    if not isinstance(data_crawl, Munch):
        # 说明不是 Munch类型的 判断是否是Dict的
        if isinstance(data_crawl, dict):
            # 如果是dict类型的,直接发
            request = data_crawl
    else:
        # 说明是munch的,那么就转成Dict的
        request = data_crawl.toDict()

    try:
        r = do_service(f"remote/dc-service/",data=request,method="post",return_type="none")
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code < 500:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/22
from api.model.Page import Page
from factdatasource.execptions import FDExecuteException
from factdatasource.persistence.jdbc import SqlWrap
from factdatasource.persistence.jdbc.JdbcFactDatasourceContext import JdbcFactDatasourceContext


class OracleJdbcFactDatasourceContext(JdbcFactDatasourceContext):

    def _size_sql(self):
        return f'SELECT COUNT(*) FROM "{self.fd().target}"'

    def _collect_sql(self, size: int):
        where = f' WHERE  rownum<={size}' if size and size > 0 else ''
        return f'SELECT * FROM "{self.fd().target}" {where}'

    def _delete_all_sql(self):
        return f'TRUNCATE TABLE "{self.fd().target}"'

    def _delete_sql(self, query: str):
        """删除数据SQL, 这里验证下这个query是否是删除语句"""
        if not query.lstrip().upper().startswith('DELETE'):
            err = f'{query}不是合法的删除操作.'
            raise FDExecuteException(err)
        return query

    def _query_page_sql(self, query: str, page: Page = None):
        """
        构造分页语句
        :param query: 原始SQL
        :param page:
        :return:
        """
        page_sql = ''
        if page:
            if page.start_row > 0:
                page_sql = 'SELECT * FROM ( '
            if page.end_row > 0:
                page_sql = f'{page_sql} SELECT TMP_PAGE.*, ROWNUM ROW_ID FROM ( '
            page_sql = f'{page_sql}{query}'
            if page.end_row > 0:
                page_sql = f'{page_sql} ) TMP_PAGE WHERE ROWNUM <= {page.end_row} '
            if page.start_row > 0:
                page_sql = f'{page_sql} ) WHERE ROW_ID > {page.start_row} '
        else:
            page_sql = query
        return page_sql

    def _generate_insert_sql(self, table, result: dict) -> SqlWrap:
        """
        构造插入语句
        :param table:
        :param result:
        :return: SqlWrap
        """
        if not result:
            return None

        keys = '('
        values = '('
        for key, value in result.items():
            # TODO 这里暂时不设置id的自增长
            # if 'id' == str(key).lower() and not value:
            #     # 如果id字段的值是空的,就忽略
            #     # 这边如果只有一个id字段这里又跳过会有问题
            #     continue

            keys = '{0}"{1}",'.format(keys, key)
            values = '{0}:{1},'.format(values, key)

        if len(keys) > 1:
            keys = keys.rstrip(',') + ')'
            values = values.rstrip(',') + ')'
        else:
            return None

        sql = f'INSERT INTO "{table}" {keys} VALUES {values}'

        param = self._param_process(result)
        return SqlWrap(sql, param)

    def _generate_update_sql(self, table, result: dict) -> SqlWrap:
        """
        生成修改语句的SQL
        :param table:
        :param result:
        :return: SqlWrap
        """
        if not result:
            return None

        primary_keys = self.get_primary_keys()
        lower_primary_keys = [str(key).lower() for key in primary_keys]

        updates = ''
        for key, value in result.items():
            if str(key).lower() in lower_primary_keys:
                continue
            updates = '{0}"{1}"=:{2},'.format(updates, key, key)
        if not updates:
            return None
        updates = updates.rstrip(',')

        where = ''
        for key in primary_keys:
            where = '{0} "{1}"=:{2} AND'.format(where, key, key)
        where = where.rstrip('AND')

        sql = f'UPDATE "{table}" SET {updates} WHERE {where}'

        param = self._param_process(result)
        return SqlWrap(sql, param)




# -*- coding: utf-8 -*-
import sys
import logging

_LOG_LEVEL_ = "INFO"


def set_level(level):
    global _LOG_LEVEL_
    lv = level.upper()
    if lv in ("DEBUG", "INFO", "WARN", "ERROR"):
        _logger_ = get_logger("Logger", "INFO")
        _logger_.info("global logger set to %s" % lv)
        _LOG_LEVEL_ = lv


def get_logger(tag="py_eureka_client", level=None):
    logger = logging.getLogger(tag)

    _formatter_ = logging.Formatter(fmt='[%(asctime)s]-[%(name)s]-[line:%(lineno)d] -%(levelname)-4s: %(message)s',
                                    datefmt='%Y-%m-%d %H:%M:%S')
    screen_handler = logging.StreamHandler(sys.stdout)
    screen_handler.setFormatter(_formatter_)
    if level is not None:
        logger.setLevel(level.upper())
        screen_handler.setLevel(level.upper())
    else:
        logger.setLevel(_LOG_LEVEL_)
        screen_handler.setLevel(_LOG_LEVEL_)

    logger.addHandler(screen_handler)
    return logger



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/22
import json
import logging
import threading
import time

from confluent_kafka.cimpl import KafkaError, KafkaException
from munch import Munch

from api.model.FactDatasource import FactDatasource, FactDatasourceTypeEnum
from factdatasource.FactDatasourceContextHolder import FactDatasourceContextHolder
from factdatasource.dao.MultipleDatasourceHolder import get_multiple_datesource
from factdatasource.dao.kafka.MultipleKafkaDatasource import CustomKafkaClient
from libs.utils import Singleton

log = logging.getLogger('Ficus')


def wait_using_fd(fd_context_holder: FactDatasourceContextHolder, action: int, fact_datasource: FactDatasource):
    def listen_using_fd():
        fd_code = fact_datasource.code
        if fd_context_holder.is_using_fact_datasource(fd_code):
            while fd_context_holder.is_using_fact_datasource(fd_code):
                log.debug(f'监听到数据源{fd_code}的{action}事件，等待数据源使用完毕')
                time.sleep(0.5)

        if action == FactDatasourceChangeListener.ACTION_UPDATE:
            fd_context_holder.update_fact_datasource(fact_datasource)
        elif action == FactDatasourceChangeListener.ACTION_DELETE:
            fd_context_holder.remove_fact_datasource(fact_datasource)

    thread = threading.Thread(name='wait_using_fd', target=listen_using_fd, daemon=True)
    thread.start()


class FactDatasourceChangeListener(threading.Thread, Singleton):
    """
    监听数据源消息
    """
    ACTION_ADD = 0
    ACTION_UPDATE = 1
    ACTION_DELETE = 2

    FACT_DATASOURCE_CHANGE_TOPIC = 'springCloudBus'

    def __init__(self):
        threading.Thread.__init__(self)

        client: CustomKafkaClient = get_multiple_datesource(FactDatasourceTypeEnum.KAFKA).get_client()
        self.producer = client.producer
        self.consumer = client.consumer
        self.running = False
        # 守护线程运行
        self.daemon = True

    @property
    def fd_context_holder(self):
        return FactDatasourceContextHolder.instance()

    def run(self):
        if self.running:
            log.warning(f'FD监听线程已启动无需重复启动')
            return
        self.running = True

        # 这里关心已使用过的FD的修改和删除消息，添加应该不需要关心
        # 这里监听到消息有两件事， 1. fd_context的事件处理  2. 多数据源的事件处理
        # 3.修改和删除需要判断该数据源是否正在使用，如果正在使用需要等待使用完成后在进行修改
        self.consumer.subscribe([self.FACT_DATASOURCE_CHANGE_TOPIC])

        while self.running:
            try:
                msg = self.consumer.poll(timeout=1.0)
                if msg is None:
                    continue
                if msg.error():
                    if msg.error().code() == KafkaError._PARTITION_EOF:
                        continue
                    else:
                        # Error
                        raise KafkaException(msg.error())
                else:
                    # 进行消息的处理
                    message = Munch(json.loads(msg.value()))
                    # 只处理FactDatasourceChangeEvent事件
                    if message.type == 'FactDatasourceChangeEvent':
                        self.deal_message(message)
            except Exception as e:
                log.exception(f"监听FactDatasourceChange消息出现问题,", e)
            finally:
                if self.running == False:
                    # 把消息队列关闭掉
                    self.consumer.close()

    def stop(self):
        self.running = False

    def deal_message(self, change_event):
        action = change_event.action
        fact_datasource = FactDatasource(**change_event.factDatasource)
        fd_code = fact_datasource.code

        if action == self.ACTION_ADD:
            # 不关心添加数据源事件
            return

        if self.fd_context_holder.is_loaded_fact_datasource(fd_code):
            # 本地数据源还没有使用不进行处理
            return

        if self.fd_context_holder.is_using_fact_datasource(fd_code):
            wait_num = 0
            while self.fd_context_holder.is_using_fact_datasource(fd_code):
                log.debug(f'监听到数据源{fd_code}的{action}事件，等待数据源使用完毕')
                wait_num += 1
                time.sleep(0.5)
                # 5秒都还在使用的话就只能使用新线程来处理了,否则无法处理后续的事件
                if wait_num > 10:
                    wait_using_fd(self.fd_context_holder, action, fact_datasource)
                    return

        if action == self.ACTION_UPDATE:
            self.fd_context_holder.update_fact_datasource(fact_datasource)
        elif action == self.ACTION_DELETE:
            self.fd_context_holder.remove_fact_datasource(fact_datasource)



import threading

from confluent_kafka.cimpl import Consumer, KafkaError, KafkaException

from api.exceptions import IllegalArgumentException
from api.handler.message.IMessageHandler import IMessageHandler
from api.model.ResultVO import ResultVO, FAIL_CODE, SUCCESS
from config.annotation import Value
from factdatasource import FactDatasourceProxyService
from schedule.utils.log import FrameworkHandlerLogger



class IMessageCE(IMessageHandler):
    """

    """

    @Value("${actor.name:unknown}")
    def actor_name(self):
        pass

    def init_message_ce(self, site, projectCode, code) -> ResultVO:
        """
        初始化消息CE
        :param site:
        :param projectCode:
        :param code:
        :return:
        """
        from client import ComputeExecutionClient
        dataComputeExecution = ComputeExecutionClient.get(site, projectCode, code)

        if dataComputeExecution is None:
            return ResultVO(FAIL_CODE, f"执行失败,没有找到Code:{code}的Ce")

        self.dataComputeExecution = dataComputeExecution

        # 找到kafka的 来源
        sourceFdCodes = dataComputeExecution.sourceFdCodes
        if sourceFdCodes is None or len(sourceFdCodes) == 0:
            return ResultVO(FAIL_CODE, f"执行失败,没有找到Code:{code}的Ce 没有配置sourceFactData")

        sourceFd = None

        for sourceFdCode in sourceFdCodes:
            # 找到FD的定义
            fd = FactDatasourceProxyService.fd_client_proxy().fd(sourceFdCode)
            if fd is None:
                return ResultVO(FAIL_CODE, f"执行失败,没有找到Code:{code}的Ce 配置的sourceFactData {sourceFdCode} 不存在")
            if fd.type in ["KAFKA", "AMQP", "JMS"]:
                # 说明找到了kafka或amqp或jms的来源fd,只找头一个
                sourceFd = fd
                break

        if sourceFd is None:
            return ResultVO(FAIL_CODE, f"执行失败,没有找到Code:{code}的Ce 没有配置消息的来源")

        # 到这里来fd只可能是两种: KAFKA或AMQP或jms
        # 现在先只支持kafka

        # 启动消息队列的监听
        self.t = MessageListener(self, f"{site}.{projectCode}.{code}", sourceFd.target, sourceFd.connection)
        self.t.setDaemon(False)  # 设置为非守护线程.
        self.t.start()

        return SUCCESS

    def check_and_init(self):
        """
        定期执行，检测是否初始化，如果没有则初始化
        :return:
        """
        if self.is_initialed():
            return

        # 说明ficus都没有启动,没法始化消息监听,放弃
        if not self.check_ficus_online():
            FrameworkHandlerLogger.log_error("启动时,没有可用的ficus服务,忽略初始化消息监听")
            return

        self.set_initialed()

        # region 找到自己这个类的名字和对象,如果是空的就放弃
        from api.annotation.annotation import TASK_HANDLERS
        if len(TASK_HANDLERS) == 0:
            return

        task_handler_name = None
        for k, v in TASK_HANDLERS.items():
            if v is self:
                task_handler_name = k
                break

        if task_handler_name is None:
            return
        # endregion

        FrameworkHandlerLogger.log_info("MessageCE开始初始化消息队列")

        from registry import REGISTRY_PROPERTIES
        registry = REGISTRY_PROPERTIES

        if registry is None or len(registry) == 0 or "computeexecutions" not in registry:
            # 说明没有yml文件,不进行注册处理
            FrameworkHandlerLogger.log_warn("MessageCE开始初始化消息队列,registry.yml中没有配置ce 不自动初始化.请使用调度触发")
            return

        # 遍历,找到这个CE的配置
        index = 1
        for computeexecution in registry["computeexecutions"]:
            if "handler" not in computeexecution or computeexecution["handler"] == "":
                raise IllegalArgumentException(f"初始化消息队列{computeexecution['code']}失败,执行器为空")

            if task_handler_name == computeexecution["handler"]:
                # 说明是一样的ce
                # 那么这里就需要开始进行初始化

                if "code" not in computeexecution or computeexecution["code"] == "" or computeexecution["code"] is None:
                    computeexecution["code"] = f"{self.actor_name()}-{index}"

                code = computeexecution["code"]
                site = registry["site"]
                dataproject = registry["dataproject"]

                stringResultVO = self.init_message_ce(site, dataproject, code)

                FrameworkHandlerLogger.log_info(
                    "MessageCE开始初始化消息队列,初始化CE:" + site + "_" + dataproject + "_" + code + " 完成,结果:" + stringResultVO.msg)

                from api.model.ResultVO import SUCCESS_CODE
                if stringResultVO.code == SUCCESS_CODE:
                    # 成功一个就不继续了
                    self.set_params({"site_": site, "projectCode_": dataproject, "code_": code})
                    break

            index = index + 1

    def whileCheck(self):
        while self.is_initialed() == False:
            self.check_and_init()
            import time
            time.sleep(30)

    def __init__(self):
        IMessageHandler.__init__(self)

        # 这里启动一个线程来检测是否启动了
        self.mthread = threading.Thread(target=self.whileCheck)
        self.mthread.setDaemon(False)
        self.mthread.start()

        self.t = None

    def stop(self):
        if self.t is not None:
            self.t.stop()

        self.initialed = False


class MessageListener(threading.Thread):
    """
        消息队列的处理器
    """

    def __init__(self, messageCe: IMessageCE, group_id, topic, brokers):
        threading.Thread.__init__(self)
        # 初始化
        self.messageCe = messageCe

        # 开启kafka的消息监听
        config = {'bootstrap.servers': brokers, 'group.id': group_id, 'session.timeout.ms': 6000,
                  'default.topic.config': {'auto.offset.reset': 'largest'}}
        import logging
        self.consumer = Consumer(config, logger=logging.getLogger("Ficus"))  # TODO 这里的入参好像有问题
        self.consumer.subscribe([topic])

        self.go_on = True

    def stop(self):
        self.go_on = False

    def run(self):
        """
        监听消息
        :return:
        """

        # 开始循环接收消息
        while self.go_on:
            try:
                msg = self.consumer.poll(timeout=1.0)
                if msg is None:
                    continue
                if msg.error():
                    if msg.error().code() == KafkaError._PARTITION_EOF:
                        continue
                    else:
                        # Error
                        raise KafkaException(msg.error())
                else:
                    # 进行消息的处理
                    self.messageCe.message(msg.value())
            except Exception as e:
                FrameworkHandlerLogger.log_error(f"接收消息出现问题,", e)
            finally:
                if self.go_on == False:
                    # 把消息队列关闭掉
                    self.consumer.close()



import threading

__context_holder = threading.local()

class Sharding(object):
    index = -1
    total = -1

    def __init__(self, index=2, total=-1):
        self.index = index
        self.total = total

    def __str__(self) -> str:
        return str({"index":self.index,"total":self.total})


def set_sharding(sharding:Sharding):
    __context_holder.key = sharding

def get_sharding() -> Sharding:
    return __context_holder.key

def reset():
    __context_holder.key = None


# 日志文件写入器

# 本地log
import logging
import os
import threading
from datetime import datetime

from munch import Munch

local_log = logging.getLogger('Ficus')

# TODO 这里可能有问题,python中还没找到能在子线程中继承的threadLocal,也就是类似JAVA中InheritableThreadLocal的东西
__context_holder = threading.local()

__log_base_path = "/sobeyHiveLogs/sdcLog/JOB/"

def init_log_path(log_path: str):
    global __log_base_path
    if log_path is not None and log_path.strip() != "":
        __log_base_path = log_path

    # 创建文件夹
    if not os.path.exists(__log_base_path):
        try:
            os.makedirs(__log_base_path, exist_ok=True)
        except Exception as e:
            local_log.error(f"初始化日志路径:{__log_base_path} 失败,{str(e)}")
            raise e


def get_log_path() -> str:
    return __log_base_path


def prepare_to_log(trigger_date: datetime, log_id):
    __context_holder.key = _make_log_file_name(trigger_date, log_id)


def end_log():
    __context_holder.key = None


def get_log_file_name() -> str:
    try:
        return __context_holder.key
    except :
        return None

def get_log_file_path(trigger_date: datetime, log_id) -> str:
    return os.path.join(get_log_path(), trigger_date.strftime("%Y-%m-%d"), (str(log_id) if log_id is not None else "0")+".log")


def _make_log_file_name(trigger_date: datetime, log_id):
    dir_path = os.path.join(get_log_path(), trigger_date.strftime("%Y-%m-%d"))

    if not os.path.exists(dir_path):
        try:
            os.mkdir(dir_path)
        except OSError as e:
            import errno
            if e.errno != errno.EEXIST:
                local_log.error(f"创建日志路径{dir_path}失败,{str(e)}")
                raise

    #  filePath/yyyy-MM-dd/9999.log
    return os.path.join(dir_path, (str(log_id) if log_id is not None else "0")+".log")


def append_log(log_file_name: str, append_log: str):
    """
    追加方式写文件
    :param log_file_name:
    :param append_log:
    :return:
    """
    if log_file_name is None or log_file_name.strip() == "":
        return

    append_log = append_log if append_log is not None else ""

    with open(log_file_name,"a+",encoding="utf-8") as file:
        try:
            file.writelines([append_log,"\n"])
        except Exception as e:
            local_log.error(f"写文件失败,{str(e)}")



def read_log(instance_address:str,trigger_date: datetime, log_id, from_line_num: int, to_line_num: int)->Munch:
    """
    读取文件内容
    :param trigger_date:
    :param log_id:
    :param from_line_num:
    :param to_line_num:
    :return:
    """
    log_file_path = get_log_file_path(trigger_date,log_id)

    if not os.path.exists(log_file_path):
        munch = Munch()
        munch.fromLineNum = from_line_num
        munch.toLineNum = to_line_num
        munch.instanceAddress = instance_address
        munch.content = f"无日志输出或日志文件已清除({log_file_path})"
        munch.end = True
        return munch

    # 这个是文件存在
    with open(log_file_path,"rt",encoding="utf-8") as file:
        logContentBuffer = []
        _toLineNum = 0
        # index 0开始
        for index, content in enumerate(file):
            _toLineNum = index

            if to_line_num >0 and _toLineNum >= to_line_num:
                # 超过了目标,就直接返回了
                munch = Munch()
                munch.fromLineNum = from_line_num
                munch.toLineNum = to_line_num
                munch.content = ''.join(logContentBuffer)
                munch.instanceAddress = instance_address
                munch.end = False
                return munch

            if _toLineNum >= from_line_num:
                logContentBuffer.append(content)

        content = ''.join(logContentBuffer)
        if content.strip()=="":
            content = "文本内容为空"

        munch = Munch()
        munch.fromLineNum = from_line_num
        munch.toLineNum = _toLineNum
        munch.content = content
        munch.end = True
        munch.instanceAddress = instance_address
        return munch


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/21
import json
import logging
from abc import abstractmethod

from munch import Munch

from api.exceptions import IllegalArgumentException
from api.model.Page import Page
from factdatasource.dao.FactDatasource import customer_dao_context_holder as costomer_context
from factdatasource.dao.jdbc.JdbcDatasourceDao import JdbcDatasourceDao
from factdatasource.execptions import FDExecuteException
from factdatasource.persistence.AbstractFactDatasourceContext import AbstractFactDatasourceContext
from factdatasource.persistence.SqlableDeleteCondition import SqlableDeleteCondition
from factdatasource.persistence.jdbc import SqlWrap
from libs.utils import str_placeholder_replace

log = logging.getLogger('Ficus')


class JdbcFactDatasourceContext(AbstractFactDatasourceContext, SqlableDeleteCondition):

    @property
    def dao(self) -> JdbcDatasourceDao:
        return JdbcDatasourceDao.instance()

    def size(self):
        """
        返回数据总长度
        :return: 数据条数:long
        """
        source_name = self.fd().get_source_name()

        sql = self._size_sql()
        log.debug(f'事实库{source_name}执行size操作SQL：{sql}')
        try:
            costomer_context.set_source(source_name)
            result = self.dao.select_num(sql)
            size = 0 if result is None else result
            return size
        except Exception as e:
            error = f'事实库{source_name}执行size操作SQL：{sql} 发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    @abstractmethod
    def _size_sql(self):
        """
        构造查询数据总条数的sql
        :return:
        """

    def is_empty(self):
        """
        返回是否存在数据
        :return: boolean
        """
        source_name = self.fd().get_source_name()

        sql = self._is_empty_sql()
        log.debug(f'事实库{source_name}执行is_empty操作SQL：{sql}')
        try:
            costomer_context.set_source(source_name)
            result = self.dao.select_num(sql)
            size = 0 if result is None else result
            return size <= 0
        except Exception as e:
            error = f'事实库{source_name}执行is_empty操作SQL：{sql} 发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def _is_empty_sql(self):
        """
        构造查询数据总条数的sql
        :return:
        """
        return self._size_sql()

    def collect(self, size: int):
        """
        返回指定条数的数据
        :param size: 返回的条数
        :return: list
        """
        source_name = self.fd().get_source_name()

        sql = self._collect_sql(size)
        log.debug(f'事实库{source_name}执行collect操作SQL：{sql}')
        try:
            costomer_context.set_source(source_name)
            result = self.dao.select_all(sql)
            return result
        except Exception as e:
            error = f'事实库{source_name}执行collect操作SQL：{sql} 发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def collect_conditions(self, size: int, condition_groups: list):
        """
        返回指定条数的数据
        :param size: 返回的条数
        :param condition_groups: 查询条件
        :return: list
        """
        source_name = self.fd().get_source_name()
        if condition_groups is None:
            sql = self._collect_sql(size)
        else:
            sql = self.search_condition_sql(self.fd().target, condition_groups)
            if size is not None and size > 0:
                # 要分页
                sql = self._query_page_sql(sql, Page(pageNum=1, pageSize=size))

        log.debug(f'事实库{source_name}执行collect操作SQL：{sql}')
        try:
            costomer_context.set_source(source_name)
            result = self.dao.select_all(sql)
            return result
        except Exception as e:
            error = f'事实库{source_name}执行collect操作SQL：{sql} 发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    @abstractmethod
    def _collect_sql(self, size: int):
        """查询指定条数数据的SQL"""

    def query(self, query: str, parameters: dict = None):
        """
        使用查询语句查询数据
        :param query: 查询语句
        :param parameters: 查询参数
        :return: Page
        """
        source_name = self.fd().get_source_name()

        page = None
        if parameters and 'pageNum_' in parameters.keys() and 'pageSize_' in parameters.keys():
            # 需要分页的
            param = {
                'pageNum': parameters.get('pageNum_'),
                'pageSize': parameters.get('pageSize_'),
                'needCount': parameters.get('needCount_') if parameters.get('needCount_') else False
            }
            page = Page(**param)

        # 参数占位符处理
        query = self._fd_placeholder_replace(query, parameters)

        sql = self._query_page_sql(query, page)

        log.debug(f'事实库{source_name}执行query操作SQL：{sql}')
        try:
            costomer_context.set_source(source_name)

            result = self.dao.select_all(sql, parameters)
            if page:
                total = None
                if page.need_count():
                    total = self.dao.select_total(query, parameters)
                page.set_result(result, total)
                return page
            else:
                return Page.no_page(result)
        except Exception as e:
            error = f'事实库{source_name}执行query操作SQL：{sql} 发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    @abstractmethod
    def _query_page_sql(self, query: str, page: Page):
        """
        构造分页语句
        :param query:
        :param page:
        :return:
        """

    def _single_thread_inserts(self, table: str, result_list: list):
        """
        批量保存数据,要求list里面的字段和数据库里面的字段一一对应
        :param table: 表名
        :param result_list: 要保存的数据
        :return:
        """
        source_name = self.fd().get_source_name()
        rrr=[]
        try:
            costomer_context.set_source(source_name)
            for result in result_list:
                self._inner_insert(table, result, 0)
                rrr.append(Munch({"success": True}))
        finally:
            costomer_context.clear_source()
        return rrr

    def _inner_insert(self, table: str, result, dept: int):
        if dept > 5:
            # 重试5次都不行，直接报错
            raise FDExecuteException(f"插入数据{result}错误，重试5次依然失败。")
        source_name = self.fd().get_source_name()

        if isinstance(result, dict):
            sql_wrap = self._generate_insert_sql(table, result)
        else:
            sql_wrap = self._generate_insert_sql(table, vars(result))

        if not sql_wrap:
            log.error(f'事实库{source_name}执行inserts操作时时无法对数据{result}进行有效的插入，放弃该数据的添加。')
            return

        source_name = self.fd().get_source_name()
        try:
            self.dao.execute(sql_wrap.sql, sql_wrap.param)
            # TODO  这里捕获的异常需要更精细的划分
        except Exception as e:
            error = f'事实库{source_name}执行inserts操作SQL：{sql_wrap.sql} 发生异常, {str(e)}'
            log.error(error)
            # 重试
            self._inner_insert(table, result, dept + 1)

    @abstractmethod
    def _generate_insert_sql(self, table, result: dict) -> SqlWrap:
        """
        生成插入语句的SQL
        :param table:
        :param result:
        :return: SqlWrap
        """

    def _single_thread_updates(self, table: str, result_list: list):
        """
        批量更新数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要修改的数据
        :return:
        """
        source_name = self.fd().get_source_name()
        rrr=[]
        try:
            costomer_context.set_source(source_name)
            for result in result_list:
                self._inner_update(table, result)
                rrr.append(Munch({"success": True}))
        finally:
            costomer_context.clear_source()
        return rrr

    def _inner_update(self, table, result) -> bool:
        if isinstance(result, dict):
            sql_wrap = self._generate_update_sql(table, result)
        else:
            sql_wrap = self._generate_update_sql(table, vars(result))

        source_name = self.fd().get_source_name()

        if not sql_wrap:
            log.error(f'事实库{source_name}执行updates操作时时无法对数据{result}进行有效的更新，放弃该数据的更新。')
            return False

        try:
            return self.dao.execute(sql_wrap.sql, sql_wrap.param).rowcount > 0
        except Exception as e:
            error = f'事实库{source_name}执行updates操作SQL：{sql_wrap.sql} 发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        return False

    @abstractmethod
    def _generate_update_sql(self, table, result: dict) -> SqlWrap:
        """
        生成修改语句的SQL
        :param table:
        :param result:
        :return: SqlWrap
        """

    def _single_thread_inserts_or_updates(self, table: str, result_list: list):
        """
        批量saveOrUpdate数据,,数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要添加或者需要修改的数据
        :return:
        """
        source_name = self.fd().get_source_name()
        rrr=[]
        try:
            costomer_context.set_source(source_name)
            for result in result_list:
                if not self._inner_update(table, result):
                    # 没有更新成功
                    try:
                        self._inner_insert(table, result, 0)
                    except FDExecuteException as e:
                        # 这种异常直接抛出去
                        raise e
                    except Exception as e:
                        # 如果插入失败了,就可能是已经有其他的线程插入了,那么就重新执行更新.这里只尝试一次,而不是做递归
                        self._inner_update(table, result)
                rrr.append(Munch({"success": True}))
        finally:
            costomer_context.clear_source()
        return rrr

    def delete_all(self):
        """
        清空数据
        :return:
        """
        source_name = self.fd().get_source_name()

        sql = self._delete_all_sql()
        log.debug(f'事实库{source_name}执行delete_all操作SQL：{sql}')
        try:
            costomer_context.set_source(source_name)
            self.dao.execute(sql)
        except Exception as e:
            error = f'事实库{source_name}执行delete_all操作SQL：{sql} 发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    @abstractmethod
    def _delete_all_sql(self):
        """清空数据SQL"""

    def delete(self, query: str):
        """
        根据删除语句删除数据,query是完整的删除语句
        :param query: 完整的删除语句
        :return:
        """
        source_name = self.fd().get_source_name()

        sql = self._delete_sql(query)
        log.debug(f'事实库{source_name}执行delete操作SQL：{sql}')
        try:
            costomer_context.set_source(source_name)
            self.dao.execute(sql)
        except Exception as e:
            error = f'事实库{source_name}执行delete操作SQL：{sql} 发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def delete_conditions(self, condition_groups: list):
        """
        根据删除条件,构造删除语句
        :param condition_groups:
        :return:
        """
        source_name = self.fd().get_source_name()
        sql = self.delete_condition_sql(self.fd().target, condition_groups)
        log.debug(f'事实库{source_name}执行delete操作SQL：{sql}')
        try:
            costomer_context.set_source(source_name)
            self.dao.execute(sql)
        except Exception as e:
            error = f'事实库{source_name}执行delete操作SQL：{sql} 发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    @abstractmethod
    def _delete_sql(self, query: str):
        """删除数据SQL"""

    def _fd_placeholder_replace(self, value: str, value_map: dict):
        """
        处理JDBC中的参数占位符
        将${key}这种直接进行替换， 将#{key}更换为:key这种，然后通过传参到数据库中去
        :param value:
        :param value_map:
        :return:
        """
        if not isinstance(value, str):
            raise IllegalArgumentException('占位符处理参数类型错误，value只能为str.')
        value = str_placeholder_replace('${', '}', value, value_map)

        tmp_map = {}
        for key in value_map.keys():
            tmp_map[key] = ':' + str(key)
        value = str_placeholder_replace('#{', '}', value, tmp_map)
        return value

    def _param_process(self, param: dict):
        """
        构造sql时处理下SQL的传入参数
        主要用于将对象字符串化，方便系统构造sql
        :param param:
        :return:
        """
        result = {}

        if not param:
            result

        for key, value in param.items():
            if isinstance(value, str):
                v = value
            elif isinstance(value, list) or isinstance(value, dict):
                v = json.dumps(value, ensure_ascii=False)
            elif isinstance(value, tuple) or isinstance(value, set):
                v = str(value)
            else:
                v = value
            result[key] = v
        return result



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/3/8
import threading
from abc import abstractmethod
from datetime import datetime
from unittest.mock import patch, Mock

from munch import Munch

from api.exceptions import IllegalArgumentException
from api.handler.ICacheAbleHandler import ICacheAbleHandler
from api.handler.ITaskHandler import ITaskHandler
from api.handler.ce.AbstractBatchCE import AbstractBatchCE
from api.handler.ce.ISimpleScriptCE import ISimpleScriptCE
from api.handler.outputer.ISimpleOutputer import ISimpleOutputer
from api.handler.script import ScriptHandlerHolder
from api.model.FdInputPipe import FdInputPipe
from api.model.ResultVO import ResultVO, FAIL_CODE, SUCCESS
from client import ComputeExecutionClient
from schedule.test.suite import BaseInitTestSuite, PROCESS_LOG_ID, LOG_ID, ScheduleCacheClientMock, TaskLoggerMock
from schedule.test.suite import DataComputeExecution


class InnerScriptPythonCETestHandler(ITaskHandler, ISimpleOutputer, ICacheAbleHandler):

    def __init__(self, simple_script_ce: ISimpleScriptCE):
        super().__init__()
        self._simple_script_ce = simple_script_ce
        self.__code_local_host = threading.local()
        self.__process_id_local = threading.local()
        self.__execution_message_local = threading.local()

    def execute(self, params: dict):
        if params is None or len(params) == 0 or ("site_" not in params) or ("projectCode_" not in params) or (
                "code_" not in params):
            # 不存在ce的信息,没法执行
            return ResultVO(FAIL_CODE, "执行失败,没有ce的信息")

        # 找到对应的dataCrawl
        data_ce = ComputeExecutionClient.get(params["site_"], params["projectCode_"], params["code_"])

        if data_ce is None or data_ce.type != "CUSTOM":
            return ResultVO(FAIL_CODE, f"执行失败,没有找到Code:{params['site_']}的ce,或者该ce类型不为Custom")

        # 赋值 日志的上下文
        # from schedule.utils.log.TaskLogger import TaskLogger
        # from schedule.utils.log import TaskLogFileAppender
        # log_path = TaskLogFileAppender.prepare_to_log(datetime.now(), 0)
        # self._simple_script_ce.task_logger = TaskLogger(log_path)
        # 这里日志输出到控制台，不向日志文件中中记录
        self._simple_script_ce.task_logger = TaskLoggerMock()

        try:
            self.set_local_code(data_ce.site + "_" + data_ce.projectCode + "_" + data_ce.code)
            self.set_process_id(PROCESS_LOG_ID)
            ScriptHandlerHolder.holder.key = self
            self.__execution_message_local.content = []

            result_list = self._simple_script_ce.do_compute(FdInputPipe(data_ce.sourceFdCodes), params)
        except Exception as e:
            return ResultVO(FAIL_CODE, f"执行失败,Code:{params['site_']}的CE,发生错误:{str(e)}")
        finally:
            # 清理 MessageLocal 和 LocalCode
            self.clear_local_code()
            self.clear_process_id()
            self.__execution_message_local.content = None
            ScriptHandlerHolder.holder.key = None
            self._simple_script_ce.task_logger = None
        return self._send_results(params, result_list, data_ce.outputFdCodes)

    def _send_results(self, params, result_list, output_fd_codes):
        """
        发送结果数据
        :param params:
        :param result_list:
        :param output_fd_codes:
        :return:
        """
        if result_list is None or len(result_list) == 0:
            # 搞完了,没的结果,不处理
            return SUCCESS
        # 有结果,就需要从ce的配置中找到目标的fd,然后调用fd的接口进行保存
        try:
            self.send_output_result(params["code_"], result_list, output_fd_codes)
        except Exception as e:
            self._simple_script_ce.task_logger(
                f"{params['site_']}_{params['projectCode_']}_{params['code_']} 发送结果数据失败,", e)
            return ResultVO(FAIL_CODE, f"发送结果数据失败,错误:{str(e)}")
        return SUCCESS

    def get_code_thread_local(self):
        return self.__code_local_host

    def get_process_thread_local(self):
        return self.__process_id_local

    def get_execution_message_cache(self):
        return self.__execution_message_local.content


class AbstractSimpleScriptCETestSuite(BaseInitTestSuite):
    """
    CeSimple脚本测试组件
    """
    # region 需要实现的方法
    @abstractmethod
    def mock_target(self) -> ISimpleScriptCE:
        """
        设置ce实例
        :return:
        """

    @abstractmethod
    def mock_fact_datasource(self) -> dict:
        """
        设置FD的基本定义,不区分数据源和主数据
        :return: dict<code,FactDatasource>
        """

    @abstractmethod
    def mock_ce(self) -> DataComputeExecution:
        """
        设置ce的基本定义
        :return: DataComputeExecution
        """

    def mock_cache_from_process(self) -> dict:
        return {}
    # endregion

    # region 内部逻辑
    def setUp(self):
        target = self.mock_target()
        if not target:
            raise IllegalArgumentException('测试抓取器实例为空')

        fds = self.mock_fact_datasource()
        if not fds:
            raise IllegalArgumentException('测试数据源为空')

        ce = self.mock_ce()
        if not ce:
            raise IllegalArgumentException('测试抓取器定义为空')

    def __fd_client(self):
        return self._get_datesource_client()

    def __mock_ce(self):
        ce = self.mock_ce()
        return Munch(ce.to_dict())

    @patch('schedule.utils.log.TaskLogFileAppender.prepare_to_log')
    @patch('factdatasource.FactDatasourceProxyService.fd_client_proxy')
    @patch('client.ComputeExecutionClient.get')
    def ce_execute(self, mock_ce_get, mock_fd_client, mock_prepare_to_log):
        """
        执行测试
        :return:
        """
        mock_ce_get.return_value = self.__mock_ce()
        mock_fd_client.return_value = self.__fd_client()
        mock_prepare_to_log.return_value = None

        ce = self.mock_ce()
        ce_param = {
            'site_': ce.site,
            'projectCode_': ce.projectCode,
            'code_': ce.code
        }
        ce_param.update(ce.params)

        target: ISimpleScriptCE = self.mock_target()
        ce_handler = InnerScriptPythonCETestHandler(target)

        if self.distributed_mode():
            # 模拟缓存
            cache_mock = ScheduleCacheClientMock()
            with patch('client.ScheduleCacheClient.set_value', Mock(side_effect=cache_mock.set_value)), \
                 patch('client.ScheduleCacheClient.set_if_absent', Mock(side_effect=cache_mock.set_if_absent)), \
                 patch('client.ScheduleCacheClient.get', Mock(side_effect=cache_mock.get)), \
                 patch('client.ScheduleCacheClient.delete', Mock(side_effect=cache_mock.delete)):
                self.execute(ce_handler, ce_param)
        else:
            self.execute(ce_handler, ce_param)

    def test_ce_execute(self):
        """
        测试入口
        :return:
        """
        self.ce_execute()
    # endregion


class AbstractBatchCETestSuite(BaseInitTestSuite):
    """
    CeSimple脚本测试组件
    """
    # region 需要实现的方法
    @abstractmethod
    def mock_target(self) -> AbstractBatchCE:
        """
        设置ce实例
        :return:
        """

    @abstractmethod
    def mock_fact_datasource(self) -> dict:
        """
        设置FD的基本定义,不区分数据源和主数据
        :return: dict<code,FactDatasource>
        """

    @abstractmethod
    def mock_ce(self) -> DataComputeExecution:
        """
        设置ce的基本定义
        :return: DataComputeExecution
        """

    def mock_cache_from_process(self) -> dict:
        return {}

    def mock_cache_from_task(self) -> dict:
        return {}
    # endregion

    # region 内部逻辑
    def setUp(self):
        target = self.mock_target()
        if not target:
            raise IllegalArgumentException('测试抓取器实例为空')

        fds = self.mock_fact_datasource()
        if not fds:
            raise IllegalArgumentException('测试数据源为空')

        ce = self.mock_ce()
        if not ce:
            raise IllegalArgumentException('测试抓取器定义为空')

    def __fd_client(self):
        return self._get_datesource_client()

    def __mock_ce(self):
        ce = self.mock_ce()
        return Munch(ce.to_dict())

    @patch('schedule.utils.log.TaskLogFileAppender.prepare_to_log')
    @patch('factdatasource.FactDatasourceProxyService.fd_client_proxy')
    @patch('client.ComputeExecutionClient.get')
    def ce_execute(self, mock_ce_get, mock_fd_client, mock_prepare_to_log):
        """
        执行测试
        :return:
        """
        mock_ce_get.return_value = self.__mock_ce()
        mock_fd_client.return_value = self.__fd_client()
        mock_prepare_to_log.return_value = None

        ce = self.mock_ce()
        ce_param = {
            'site_': ce.site,
            'projectCode_': ce.projectCode,
            'code_': ce.code,
            '__logId__': LOG_ID,
            '__triggerTime__': datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            '__processLogId__': PROCESS_LOG_ID
        }
        ce_param.update(ce.params)

        target: AbstractBatchCE = self.mock_target()

        # 这里日志输出到控制台，不向日志文件中中记录
        target.task_logger = TaskLoggerMock()

        target.set_local_code(f"{ce.site}_{ce.projectCode}_{ce.code}")
        target.set_process_id(PROCESS_LOG_ID)

        if self.distributed_mode():
            # 模拟缓存
            cache_mock = ScheduleCacheClientMock()
            with patch('client.ScheduleCacheClient.set_value', Mock(side_effect=cache_mock.set_value)), \
                 patch('client.ScheduleCacheClient.set_if_absent', Mock(side_effect=cache_mock.set_if_absent)), \
                 patch('client.ScheduleCacheClient.get', Mock(side_effect=cache_mock.get)), \
                 patch('client.ScheduleCacheClient.delete', Mock(side_effect=cache_mock.delete)):

                for key, value in self.mock_cache_from_process().items():
                    target.set_cache_value_from_process(key,value)

                for key, value in self.mock_cache_from_task().items():
                    target.set_cache_value(key,value)

                self.execute(target, ce_param)
        else:
            self.execute(target, ce_param)

    def test_ce_execute(self):
        """
        测试入口
        :return:
        """
        self.ce_execute()
    # endregion



import logging
from typing import Any

from rule_engine import RuleFacts, RuleConditionRelationEnum
from rule_engine.core.SimpleValueProvider import SimpleValueProvider

log = logging.getLogger('Ficus')

def match(conditions:list,facts:RuleFacts) -> bool:
    """
    条件的比对
    :param conditions:
    :param facts:
    :return:
    """
    global factClass
    matchResult = True

    # 遍历每一个规则因子
    for condition in conditions:
        if not condition.enable:
            log.info(f"规则因子:{condition.source}未启用,直接跳过")
            continue

        relation = condition.relation

        if condition.subs is not None and len(condition.subs)>0:
            # 有子条件的,做递归
            matchResult = relation.relation(matchResult,match(condition.subs,facts))
            if relation == RuleConditionRelationEnum.AND and not matchResult:
                # 如果关系是AND,并且结果是失败,那就直接返回了
                return False
        else:
            # 无子条件的
            # 2.找到规则因子的名字.这个名字就是事实的名字
            # 3.找到该规则因子的事实值
            factValue = facts.get_facts() if condition.source is None or condition.source.strip()=="" else facts.get_fact(condition.source)
            target = None
            if condition.target_provider is None:
                # 如果没有自定义targetProvider,那么就是用默认的
                # 这里需要猜测target的类型
                classes = condition.operation.allowed_target_type()

                if len(classes)==0:
                    target = SimpleValueProvider().get_value(condition.target)

                # 说明是有类型的
                if factValue is None:
                    # 没有入参,或者入参 是空,没法猜
                    target = SimpleValueProvider().get_value(condition.target)
                else:
                    # 如果输入是数组或集合(list,set,dict)
                    if hasattr(factValue,"__iter__") and not isinstance(factValue,dict):
                        if len(factValue)==0:
                            # 没法取到入参的类型
                            target = SimpleValueProvider().get_value(condition.target)
                        else:
                            factClass = type(next(iter(factValue)))
                    else:
                        factClass = type(factValue)

                    if target is None:
                        find = False
                        for a_class in classes:
                            if issubclass(factClass,a_class):
                                target = SimpleValueProvider(a_class).get_value(condition.target)
                                find=True
                                break

                        if not find:
                            # 没有找到,也就是说类型可能不匹配操作
                            target = SimpleValueProvider().get_value(condition.target)

            else:
                target = condition.target_provider.get_value(condition.target)


            # 开始做匹配
            matchResult = relation.relation(matchResult,condition.rule_operation.do_operation(factValue,target))

            if relation == RuleConditionRelationEnum.AND and not matchResult:
                # 一个条件因子匹配不成功, 就返回false
                return False

    return True


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/21
import json
import logging

from bson import SON
from munch import Munch

from api.exceptions import IllegalArgumentException
from api.model.Page import Page
from factdatasource.dao.FactDatasource import customer_dao_context_holder as costomer_context
from factdatasource.dao.mongo.MongoDatasourceDao import MongoDatasourceDao
from factdatasource.execptions import FDExecuteException
from factdatasource.persistence.AbstractFactDatasourceContext import AbstractFactDatasourceContext

log = logging.getLogger('Ficus')


class MongoFactDatasourceContext(AbstractFactDatasourceContext):

    @property
    def dao(self) -> MongoDatasourceDao:
        return MongoDatasourceDao.instance()

    def size(self):
        """
        返回数据总长度
        :return: 数据条数:long
        """
        source_name = self.fd().get_source_name()
        target = self.fd().target
        try:
            costomer_context.set_source(source_name)
            result = self.dao.count_documents(target)
            size = 0 if result is None else result
            return size
        except Exception as e:
            error = f'事实库{source_name}执行size操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def is_empty(self):
        """
        返回是否存在数据
        :return: boolean
        """
        source_name = self.fd().get_source_name()
        target = self.fd().target
        try:
            costomer_context.set_source(source_name)
            result = self.dao.count_documents(target)
            size = 0 if result is None else result
            return size <= 0
        except Exception as e:
            error = f'事实库{source_name}执行is_empty操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def collect(self, size: int):
        """
        返回指定条数的数据
        TODO 返回结果中ObjectId暂时未进行处理
        :param size: 返回的条数
        :return: list
        """
        source_name = self.fd().get_source_name()
        target = self.fd().target
        try:
            costomer_context.set_source(source_name)
            size = size if size else 0
            result = self.dao.find(target, {}, limit=size)
            return result
        except Exception as e:
            error = f'事实库{source_name}执行collect操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def collect_conditions(self, size: int, condition_groups: list):
        """
        返回指定条数的数据
        :param size: 返回的条数
        :param condition_groups: 查询条件
        :return: list
        """
        # TODO 这里需要构造mongo的bson,python没有对象方式的构建,需要自己拼json
        return None

    def query(self, query: str, parameters: dict = None):
        """
        使用查询语句查询数据
        needCount_参数的功能暂时只有普通查询会支持,map_reduce查询不会支持分页
        :param query: 查询语句
        :param parameters: 查询参数
        :return: Page
        """
        source_name = self.fd().get_source_name()

        if not query or not isinstance(query, str):
            raise IllegalArgumentException(f'事实库{source_name}执行query操作发生异常, 查询参数query：{query} 错误。')

        source_name = self.fd().get_source_name()
        target = self.fd().target

        page = None
        if parameters and 'pageNum_' in parameters.keys() and 'pageSize_' in parameters.keys():
            # 需要分页的
            param = {
                'pageNum': parameters.get('pageNum_'),
                'pageSize': parameters.get('pageSize_'),
                'needCount': parameters.get('needCount_') if parameters.get('needCount_') else False
            }
            page = Page(**param)

        # 参数占位符处理
        query = self._fd_placeholder_replace(query, parameters)

        # 处理SQL空格
        query = query.strip()

        try:
            log.debug(f'事实库{source_name}执行query操作SQL：{query}')
            costomer_context.set_source(source_name)

            if query.startswith('{') and query.endswith('}'):

                query = json.loads(query)

                if 'map' in query and 'reduce' in query and 'mapreduce' in query:
                    # 说明是map_reduce方式的函数
                    query_result = self._execute_map_reduce(target, query)
                    return Page.no_page(query_result)
                else:
                    # TODO 普通查询功能比较单一，是否需要进行扩展， 现在只支持filter,对于sort，limit,skip等没有支持
                    page = self._execute_normal_query(target, query, page)
                    return page
            elif query.startswith('[') and query.endswith(']'):
                query = json.loads(query)
                page = self._execute_aggregate_query(target, query, page)
                return page
            else:
                raise FDExecuteException(f'事实库{source_name}执行query操作查询语句错误，SQL：{query}')
        except Exception as e:
            error = f'事实库{source_name}执行query操作SQL：{query} 发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def _single_thread_inserts(self, table: str, result_list: list):
        """
        批量保存数据,要求list里面的字段和数据库里面的字段一一对应
        :param result_list: 要保存的数据
        :return:
        """
        source_name = self.fd().get_source_name()
        rrr=[]
        try:
            costomer_context.set_source(source_name)
            self.dao.insert(table, result_list)
            rrr.append(Munch({"success": True}))
        except Exception as e:
            error = f'事实库{source_name}执行inserts操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()
        return rrr

    def _single_thread_updates(self, table: str, result_list: list):
        """
        批量更新数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要修改的数据
        :return:
        """
        source_name = self.fd().get_source_name()
        rrr=[]
        try:
            costomer_context.set_source(source_name)
            primary_keys = self.get_primary_keys()
            for result in result_list:
                if not isinstance(result, dict):
                    result = vars(result)

                primary_values = self.get_primary_values(primary_keys, result)
                if not primary_values:
                    # 没有id字段，放弃该数据
                    log.warning(f'事实库{source_name}执行updates操作时未配置主键字段,放弃该数据的更新：{result}')
                    rrr.append(Munch({"success": False, "error": f'事实库{source_name}执行updates操作时未配置主键字段,放弃该数据的更新', "content": result}))
                    continue
                self.dao.update_one(table, primary_values, result)
                rrr.append(Munch({"success": True}))
        except Exception as e:
            error = f'事实库{source_name}执行updates操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()
        return rrr

    def _single_thread_inserts_or_updates(self, table: str, result_list: list):
        """
        批量saveOrUpdate数据,数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要添加或者需要修改的数据
        :return:
        """
        source_name = self.fd().get_source_name()
        rrr=[]
        try:
            costomer_context.set_source(source_name)
            primary_keys = self.get_primary_keys()
            for result in result_list:
                if not isinstance(result, dict):
                    result = vars(result)

                primary_values = self.get_primary_values(primary_keys, result)
                if not primary_values:
                    # 没有id字段,直接尝试插入
                    self.dao.insert(table, [result])
                else:
                    self.dao.update_one(table, primary_values, result,  upsert=True)
                rrr.append(Munch({"success": True}))
        except Exception as e:
            error = f'事实库{source_name}执行updates操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()
        return rrr

    def delete_all(self):
        """
        清空数据
        :return:
        """
        source_name = self.fd().get_source_name()
        target = self.fd().target
        try:
            costomer_context.set_source(source_name)
            self.dao.drop(target)
        except Exception as e:
            error = f'事实库{source_name}执行delete_all操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def delete(self, query: str):
        """
        根据删除语句删除数据,query是mongo查询语句
        :param query:
        :return:
        """
        source_name = self.fd().get_source_name()
        target = self.fd().target
        try:
            costomer_context.set_source(source_name)
            query = json.loads(query)
            self.dao.delete_many(target, query)
        except Exception as e:
            error = f'事实库{source_name}执行delete操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def delete_conditions(self, condition_groups: list):
        """
        根据删除条件,构造删除语句
        :param condition_groups:
        :return:
        """
        # TODO 这里需要构造mongo的bson,python没有对象方式的构建,需要自己拼json

    # region 查询结构构造
    def _execute_map_reduce(self, target: str, query: dict) -> list:
        """
        map_reduce 查询构造结果
        :param target:
        :param query:
        :return:
        """
        kwargs = query.copy()

        t = kwargs.pop('mapreduce')
        target = t if t else target

        map = kwargs.pop('map')
        reduce = kwargs.pop('reduce')
        if 'out' in kwargs.keys():
            out = kwargs.pop('out')
        else:
            out = SON([('inline', 1)])
        kwargs['full_response'] = True
        query_result = self.dao.map_reduce(target, map, reduce, out, **kwargs)
        if 'results' in query_result:
            return query_result.get('results')
        elif 'result' in query_result:
            return [query_result.get('result')]

    def _execute_normal_query(self, target: str, query: dict, page: Page) -> Page:
        """
        普通查询构造结果
        :param target:
        :param query:
        :param page:
        :return:  page
        """
        # 普通查询
        kwargs = {
            'filter': query
        }
        if page:
            kwargs['skip'] = page.start_row
            kwargs['limit'] = page.pageSize
        query_result = self.dao.find(target, **kwargs)

        if page:
            total = None
            if page.need_count():
                total = self.dao.count_documents(target, filter=query)
            page.set_result(query_result, total)
        else:
            page = Page.no_page(query_result)
        return page

    def _execute_aggregate_query(self, target: str, query: list, page: Page) -> Page:
        if page:
            query.append({"$skip": page.start_row})
            query.append({"$limit": page.pageSize})

        query_result = self.dao.aggregate(target, query)

        if page:
            # TODO 这里暂时不支持count查询
            page.set_result(query_result, None)
        else:
            page = Page.no_page(query_result)
        return page

    # endregion



from schedule.TaskThread import TaskThread

__actor_port = 8011
__task_thread_cache = dict()


def registry_task_thread(task_id: int, handler, remove_old_reason: str):
    """
       注册一个任务线程
    :param task_id:
    :param handler: ITaskHandler
    :param remove_old_reason:
    :return:
    """
    # 实例化一个线程
    new_task_thread = TaskThread(handler, __actor_port)
    new_task_thread.start()  # 开启线程

    old_task_thread: TaskThread = __task_thread_cache.get(task_id)
    __task_thread_cache[task_id] = new_task_thread

    if old_task_thread is not None:
        old_task_thread.stop(remove_old_reason)

    return new_task_thread


def remove_task_thread(job_id: int, remove_old_reason: str):
    """
    删除一个任务线程
    :param job_id:
    :param remove_old_reason:
    :return:
    """
    old_task_thread: TaskThread = __task_thread_cache.pop(job_id)

    if old_task_thread is not None:
        old_task_thread.stop(remove_old_reason)
        # 还需要调用 interrupt()


def cancel_task_thread(job_id: int, task_id: int, is_executing: bool):
    """
    停止某一个任务
    :param job_id:
    :param task_id:
    :param is_executing:
    :return:
    """
    task_thread: TaskThread = __task_thread_cache.get(job_id)

    if task_thread is not None:
        if is_executing:
            # 说明任务已经开始做了,不杀掉整个Job的Thread
            task_thread.kill_doing_task(task_id)
        else:
            # 说明任务还没真正开始做,还在队列里面.
            task_thread.remove_trigger_queue(task_id)


def load_task_thread(task_id: int):
    """
    返回一个任务线程
    :param task_id:
    :return:
    """
    return __task_thread_cache.get(task_id)


def count_trigger_queue() -> dict:
    """
    返回一个 Map<Integer,Integer>
    :return:
    """
    result = {}
    for k, v in __task_thread_cache.items():
        result[k] = v.count_trigger_queue()

    return result


import threading
from abc import abstractmethod
from datetime import datetime

from api.handler.ICacheAbleHandler import ICacheAbleHandler
from api.handler.ITaskHandler import ITaskHandler
from api.handler.ce.IMessageCE import IMessageCE
from api.handler.outputer.ISimpleOutputer import ISimpleOutputer
from api.model.FdInputPipe import FdInputPipe
from api.model.ResultVO import ResultVO, FAIL_CODE, SUCCESS
from schedule.utils.log import TaskLogFileAppender, FrameworkHandlerLogger


class AbstractMessageCE(ITaskHandler, IMessageCE, ISimpleOutputer, ICacheAbleHandler):
    """
    消息类型的CE的抽象类
    """

    # 由于这个的实现有可能是 单例的,所以要使用ThreadLocal
    def __init__(self):
        super().__init__()
        self.params = None
        self.dataComputeExecution = None
        self.__code_local_host = threading.local()
        self.__process_id_local = threading.local()
        self.__task_log_id_local = threading.local()
        self.__execution_message_local = threading.local()

    def message(self, data):
        """
        实现 消息接收后的处理
        :param data:
        :return:
        """

        if self.dataComputeExecution is None:
            # 表示还没有初始化
            return

        resultList = None
        try:
            self.set_local_code(
                f"{self.dataComputeExecution.site}_{self.dataComputeExecution.projectCode}_{self.dataComputeExecution.code}")
            self.set_process_id(self.params.get("__processLogId__"))
            self.__task_log_id_local.key = self.params.get("__logId__")
            TaskLogFileAppender.prepare_to_log(datetime.strptime(self.params["__triggerTime__"], "%Y-%m-%d %H:%M:%S"),
                                               self.params["code_"])
            resultList = self.do_compute(FdInputPipe(self.dataComputeExecution.sourceFdCodes), data.decode('utf-8'),
                                         self.params)
        except Exception as e:
            FrameworkHandlerLogger.log_error(
                f"{self.dataComputeExecution.site}_{self.dataComputeExecution.projectCode}_{self.dataComputeExecution.code} 接收到消息执行失败,",
                e)
            return
        finally:
            self.clear_local_code()
            self.clear_process_id()
            self.__task_log_id_local.key = None
            TaskLogFileAppender.end_log()

        if resultList is None or len(resultList) == 0:
            # 搞完了,没的结果,不处理
            return

        # 有结果,就需要从crawl的配置中找到目标的fd,然后调用fd的接口进行保存
        outputFdCodes = self.dataComputeExecution.outputFdCodes

        try:
            self.send_output_result(self.dataComputeExecution.code, resultList, outputFdCodes)
        except Exception as e:
            FrameworkHandlerLogger.log_error(
                f"{self.dataComputeExecution.site}_{self.dataComputeExecution.projectCode}_{self.dataComputeExecution.code} 发送结果数据失败,",
                e)

    def execute(self, params):
        """
        必须的方法实现
        :param params:
        :return:
        """
        if params is None or len(params) == 0 or ("site_" not in params) or ("projectCode_" not in params) or (
                "code_" not in params):
            # 不存在crawl的信息,没法执行
            return ResultVO(FAIL_CODE, "执行失败,没有ce的信息")

        # 这里始终进行刷新赋值,是为了获取最新的ce的变动
        self.params = params

        if self.dataComputeExecution is not None:
            # 说明已经初始化过了, 就不需要再初始化了
            from client import ComputeExecutionClient
            self.dataComputeExecution = ComputeExecutionClient.get(params["site_"], params["projectCode_"],
                                                                   params["code_"])
            if self.dataComputeExecution is None:
                return ResultVO(FAIL_CODE, f"执行失败,没有找到Code:{params['code_']}的Ce")

            return SUCCESS

        return self.init_message_ce(params["site_"], params["projectCode_"], params["code_"])

    def kill(self):
        """
        处理关闭的方法
        :return:
        """
        self.stop()

    def get_execution_message_cache(self):
        """
        返回message_cache
        :return:
        """
        return self.__execution_message_local.content

    def get_code_thread_local(self):
        """
        实现上下文的code
        :return:
        """
        return self.__code_local_host

    def get_process_thread_local(self):
        """
        上下文的Id
        :return:
        """
        return self.__process_id_local

    def get_task_log_id(self):
        return self.__task_log_id_local.key

    @abstractmethod
    def do_compute(self, source_fds: FdInputPipe, value, params: dict):
        """
        真正执行数据挖掘的逻辑
        :param output_stream: 数据的输出
        :param source_fds: ce的输入
        :param params: 需要的参数
        :return: list   List<OutputWrapper<Serializable>>
        """



from flask import Response

from service import TaskQueueService
from . import remote

WAITING_TASK = "waiting_tasks"
WAITING_TASK_HELP = f"# HELP {WAITING_TASK} Number of tasks on waiting queue."
WAITING_TASK_TYPE = f"# TYPE {WAITING_TASK} gauge"
WAITING_TASK_CONTENT = f"{WAITING_TASK_HELP}\n{WAITING_TASK_TYPE}\n{WAITING_TASK} "


@remote.route('/metrics', methods=['GET'])
def metrics():
    """
    返回prometheus要求的数据指标
    我们的服务就增加一个指标即可.

    # HELP waiting_tasks Number of tasks on waiting queue.
    # TYPE waiting_tasks gauge
    waiting_tasks 10

    :return:
    """
    task_queue_proxy = TaskQueueService.task_queue_proxy()

    return Response(WAITING_TASK_CONTENT + str(task_queue_proxy.count_waiting_tasks()), mimetype='text/plain')



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/21


class SqlWrap(object):
    """
    存放sql构造结果包含sql和参数
    sql 是这种 INSERT  INTO `projectdb`(`name`,`group`,`status`) VALUES (:name, :group, :status)
    param 就是与dict对应的参数
    """

    def __init__(self, sql: str, param: dict):
        self._sql = sql
        self._param = param

    @property
    def sql(self):
        return self._sql

    @property
    def param(self):
        return self._param



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/2/26
from unittest import TestCase

from api.model.FactDatasource import FactDatasource
from factdatasource.FactDatasourceContextHolder import FactDatasourceContextHolder
from factdatasource.dao.kafka.MultipleKafkaDatasource import MultipleKafkaDatasource
from factdatasource.message.FactDatasourceMessageHanlder import FactDatasourceChangeListener, wait_using_fd


class TestFactDatasourceChangeListener(TestCase):

    def test_message(self):
        # self.send_message()

        datasource = MultipleKafkaDatasource.instance()
        url = 'localhost:9092'
        datasource.add_datasource_type(datasource.DEFAULT_KAFKA_DATASOURCE, url, None)
        # 启动FD改变事件消息监听
        listener = FactDatasourceChangeListener.instance()
        listener.daemon = False
        listener.start()

    def test_wait_using_fd(self):
        fd_context = FactDatasourceContextHolder.instance()
        action = 1
        fd_data = {
            "code": "test_source",
            "connection": "localhost:9200",
            "credentials": None,
            "model": None,
            "name": "test_source",
            "privilege": None,
            "projectCode": "TEST",
            "site": "S1",
            "tags": [
                "masterData"
            ],
            "target": "es_test",
            "ttl": 0,
            "type": "ES"
        }
        fd = FactDatasource(**fd_data)
        wait_using_fd(fd_context, action, fd)





# 用来在程序中写入日志
import datetime
import logging

import traceback

# 本地log
from schedule.utils.log import TaskLogFileAppender

local_log = logging.getLogger('Ficus')


class TaskLogger:

    def __init__(self,log_file_name) -> None:
        self._log_file_name = log_file_name

    def log(self, append_log: str):
        """
            记录日志信息
            :param append_log: 需要记录的内容
            :return:
            """
        call_info = self._get_filename_function_line(limit=2)
        # print(call_info)
        self._log_detail(call_info, append_log)

    def error(self, e: Exception):
        """
            记录异常信息
            :param e:  抛出的异常
            :return:
            """
        call_info = self._get_filename_function_line(limit=2)
        self._log_detail(call_info, str(e) if e is not None else "")

    def _log_detail(self,call_info, append_log):
        """
        把东西写入文件中
        :param call_info:
        :param append_log:
        :return:
        """
        now = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]

        format_append_log = f"{now} {call_info} {append_log if append_log is not None else ''}"

        log_file_name = self._log_file_name or TaskLogFileAppender.get_log_file_name()

        if log_file_name is not None and log_file_name.strip() != "":
            TaskLogFileAppender.append_log(log_file_name, format_append_log)
        else:
            local_log.info(format_append_log)

    def _stack_tuple_to_function_line_filename(self,stackTuple):
        ''''' stackTuple: （文件名，行号，函数名，这一行的代码）
        '''
        filename = stackTuple[0]
        linenumber = stackTuple[1]
        funcname = stackTuple[2]
        filename = filename if "<string>" != filename else "动态脚本"

        import threading
        return f"[{filename}#{funcname}]-[{linenumber}]-[{threading.current_thread().name}]"


    def _get_filename_function_line(self,limit=1):  # limit = 1 表示抽取该函数调用者的位置，注意输入到extract_stack中的limit=limit+1
        stackTuple = traceback.extract_stack(limit=limit + 1)[0]
        return self._stack_tuple_to_function_line_filename(stackTuple)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/30
from api.model.FactDatasource import FactDatasource
from factdatasource.dao.jdbc.JdbcDatasourceDao import JdbcDatasourceDao
from libs.utils import Singleton


class FactDatasourceService(Singleton):

    @property
    def dao(self):
        return JdbcDatasourceDao.instance()

    def get_fd_by_code(self, fd_code: str) -> FactDatasource:
        """
        根据code查询FD的信息
        :param fd_code: 数据源code,code具有全局唯一性，所以这里可以直接根据code查询出数据
        :return: FactDatasource
        """
        if not fd_code:
            return None

        sql = f"""SELECT id,site,projectCode,code,name,description,type,connection,credentials,target,modelid AS model,ttl,tags,privilege
            FROM `sc_factdatasource` WHERE CODE=:fd_code"""
        code_dict = {'fd_code': fd_code}
        fd = self.dao.select_one(sql, code_dict)
        if fd:
            return FactDatasource(**fd)
        return None

    def exists_fds(self, fd_codes: list) -> bool:
        """
        判断fd_codes是否全部存在
        :param fd_codes:
        :return: bool
        """
        if not fd_codes:
            return False

        sql = f'SELECT code FROM `sc_factdatasource` WHERE code IN :fd_codes'
        codes_dict = {'fd_codes': fd_codes}
        codes_result = self.dao.select_all(sql, codes_dict)
        if codes_result:
            codes_list = [re['code'] for re in codes_result]
            if len(fd_codes) == len(codes_list):
                return True
        return False




#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/21
from enum import Enum

from api.model.MetaModel import MetaModel


class FactDatasource(object):
    """
    事实库的定义
    """

    def __init__(self, site, projectCode, code, name, type, connection, target,
                 id=None, description=None, credentials=None, model=None, ttl=None, tags=None,
                 privilege=None,
                 **kwargs):
        """
        描述数据源的对象
        :param site: 事实库所属的站点,必填
        :param projectCode: 事实库所属的工程,必填
        :param code: 事实库的唯一标示,必填,要求全局code唯一,允许大小写字母加数字下划线,最大128个字符
        :param name: 事实库显示名字,必填,允许中文/英文大小写字母加数字,最大256个字符
        :param type: 事实库的类型,必填 allowableValues = "FILE,JDBC,REDIS,KAFKA,MONGO,ES,SOBEY_HIVE,GRAPH,CUSTOM,AMQP"
        :param connection: 事实库的连接字符串,必填  FILE就是文件路径 JDBC/GRAPH就是jdbcUrl等等,KAFKA/AMQP/REDIS/MONGO/ES就是服务器的IP加端口 ,
            sobeyHive 就是hivecore的地址 带端口号  CUSTOM 是 微服务的名字也就是actor.name
            REIDS:单机：ip:port  集群（英文逗号分割）： ip:port,ip:port,ip:port  哨兵（第一个表示主节点名字，英文分号分割）：MASTERNAME;哨兵IP1:PORT1;哨兵IP2:PORT2;哨兵IP3:PORT3
        :param target: 事实库的数据目标,必填,FILE文件名/MYSQL,ORACLE表名/redis(key的前缀)?/kafka队列名/es索引名/mongo Collection名/
            sobeyHive 就是目标索引名 entity/ CUSTOM是 handler的名字/AMQP exchange:routingKey/ JMS 发送消息的目标位置 https://stomp.github.io/stomp-specification-1.2.html#SEND
        :param id: 事实库在数据库中的主键,添加的时候不填
        :param description: 事实库描述/备注,允许中文/英文大小写字母加数字,最大512个字符
        :param credentials: 数据的认证的字符串,JDBC/REDIS/ES/KAFKA/AMQP 就是 用户名:密码   MONGO就是 username:password@database  sobeyHive就是 站点名  CUSTOM 没意义
        :param model: 事实库关联的模型 可选
        :param ttl: 数据过期时间
        :param tags: FD的标签,用于给fd分类,可选的.其值允许英文大小写加字母下划线
        :param privilege: 权限
        """
        self.id = id
        self.site = site
        self.projectCode = projectCode
        self.code = code
        self.name = name
        self.description = description
        self.type: FactDatasourceTypeEnum = FactDatasourceTypeEnum(type)
        self.connection = connection
        self.credentials = credentials
        self.target = target
        self.model: MetaModel = MetaModel(id=model) if model is not None else None
        self.ttl = ttl
        self.tags = tags
        self.privilege = privilege
        self.other_param = kwargs

        self._source_name = f'fd_{self.site}_{self.projectCode}_{self.code}'

    def get_source_name(self):
        return self._source_name


class FactDatasourceTypeEnum(Enum):
    FILE = 'FILE'
    JDBC = 'JDBC'
    REDIS = 'REDIS'
    KAFKA = 'KAFKA'
    MONGO = 'MONGO'
    ES = 'ES'
    SOBEY_HIVE = 'SOBEY_HIVE'
    GRAPH = 'GRAPH'
    CUSTOM = 'CUSTOM'
    AMQP = 'AMQP'
    # python中JMS类型使用的是stomp协议，如果配置的JMS类型，需要开启对应消息组件的stomp支持
    JMS = 'JMS'


class JdbcTypeEnum(Enum):
    MYSQL = 'MYSQL'
    ORACLE = 'ORACLE'
    VERNOX = 'VERNOX'
    SQL_SERVER = 'SQL_SERVER'
    POSTGRESQL = 'POSTGRESQL'
    DB2 = 'DB2'
    JTDS = 'JTDS'
    MOCK = 'MOCK'
    HSQL = 'HSQL'
    SYBASE = 'SYBASE'
    ALI_ORACLE = 'ALI_ORACLE'
    MARIADB = 'MARIADB'
    DERBY = 'DERBY'
    HBASE = 'HBASE'
    HIVE = 'HIVE'
    H2 = 'H2'
    DM = 'DM'
    KINGBASE = 'KINGBASE'
    # 阿里云odps
    ODPS = 'ODPS'
    # Log4JDBC
    LOG4JDBC = 'LOG4JDBC'


class JdbcDriverEnum(Enum):
    MYSQL_DRIVER = 'mysql+mysqldb'
    ORACLE_DRIVER = 'oracle+cx_oracle'
    SQL_SERVER_DRIVER = 'mssql+pyodbc'
    POSTGRESQL_DRIVER = 'postgresql+psycopg2'
    VERNOX_DRIVER = 'mysql+mysqldb'



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/22
from confluent_kafka.cimpl import Producer, Consumer

from api.exceptions import IllegalArgumentException
from api.model.FactDatasource import FactDatasourceTypeEnum
from config.annotation import Value
from factdatasource.dao.FactDatasource import DatasourceListener, MultipleBaseDatasource, BaseDatasource, \
    customer_dao_context_holder
from factdatasource.execptions import DatasourceNotFoundException, FDExecuteException
from libs.utils import Singleton


class CustomKafkaClient(object):

    def __init__(self, producer: Producer = None, consumer: Consumer = None):
        if not producer and not consumer:
            raise IllegalArgumentException('参数不能同时为空')
        self._producer_ = producer
        self._consumer_ = consumer

    @property
    def producer(self) -> Producer:
        return self._producer_

    @property
    def consumer(self) -> Consumer:
        return self._consumer_


class MultipleKafkaDatasource(DatasourceListener, MultipleBaseDatasource, Singleton):
    """
    管理fd中的所有KAFKA类型的数据源
    """

    DEFAULT_KAFKA_DATASOURCE = '_ficus4py_default_kafka_datasource'

    def __init__(self):
        # source_name --> KafkaDatasource
        self._target_dataSources = dict()

    def get_datasource_type(self):
        """
        获取数据源类型
        :return:
        """
        return FactDatasourceTypeEnum.KAFKA

    def add_datasource_type(self, source_name: str, url: str, credentials: str):
        """
        添加一个数据源
        :param source_name:
        :param url:
        :param credentials:
        :return:
        """
        if not source_name or not url:
            raise IllegalArgumentException(f'添加数据源参数错误:source_name、url都不能为空')
        # 由于每个kafka的fd中group.id都不一样，所以kafka数据源不能多个fd共享
        # 开始创建数据源
        kafka_data_source = KafkaDatasource(source_name, url, credentials)
        kafka_data_source.start()
        self._target_dataSources[source_name] = kafka_data_source

    def update_datasource_type(self, source_name: str, url: str, credentials: str):
        """
        修改一个数据源
        :param source_name:
        :param url:
        :param credentials:
        :return:
        """
        if not source_name:
            raise IllegalArgumentException(f'修改数据源参数错误:source_name不能为空')

        self.delete_datasource_type(source_name)
        self.add_datasource_type(source_name, url, credentials)

    def delete_datasource_type(self, source_name: str):
        """
        删除一个数据源
        :param source_name:
        :return:
        """
        if not source_name:
            raise IllegalArgumentException(f'删除数据源参数错误:source_name不能为空')

        target: KafkaDatasource = self._target_dataSources.pop(source_name)
        if target:
            target.close_client()

    def get_data_source(self):
        """
        获取数据源的基本信息
        :return:
        """
        source_name = self.determine_current_lookup_key()
        if not source_name:
            raise FDExecuteException('未设置操作源,无法获取数据源信息。')

        if source_name in self._target_dataSources.keys():
            return self._target_dataSources[source_name]
        raise DatasourceNotFoundException('未发现数据源%s。' % source_name)

    def determine_current_lookup_key(self):
        """KAFKA中需要设置默认的数据源，所以这里需要重写该方法"""
        source_name = customer_dao_context_holder.get_source()
        if source_name is None:
            return self.DEFAULT_KAFKA_DATASOURCE
        return source_name

    def get_client(self):
        """
        获取数据库操作的client
        :return:
        """
        target: KafkaDatasource = self.get_data_source()
        return target.get_client()

    def close_client(self):
        """
        关闭客户端，关闭连接
        :return:
        """
        target: KafkaDatasource = self.get_data_source()
        return target.close_client()


class KafkaDatasource(BaseDatasource):

    def __init__(self, source_name, url, credentials):
        self.init(source_name, url, credentials)
        self._group_id = source_name
        self.client: CustomKafkaClient = None

    def start(self):
        """
        TODO kafka还没有支持密码验证
        :return:
        """
        producer_conf = {
            'bootstrap.servers': self.url
        }
        producer = Producer(producer_conf)

        consumer_conf = {
            'bootstrap.servers': self.url,
            'group.id': self._group_id
        }
        consumer = Consumer(consumer_conf)

        self.client = CustomKafkaClient(producer, consumer)

    def get_client(self) -> CustomKafkaClient:
        return self.client

    def close_client(self):
        if self.client and self.client.consumer:
            self.client.consumer.close()


@Value("${kafka.ip}")
def kafka_ip():
    pass


@Value("${kafka.timeout}")
def kafka_timeout():
    pass


def register_kafka_default_datasource():
    """
    注册默认的数据源
    :return:
    """
    datasource = MultipleKafkaDatasource.instance()
    url = kafka_ip()
    datasource.add_datasource_type(datasource.DEFAULT_KAFKA_DATASOURCE, url, None)










import warnings
from abc import abstractmethod

from api.handler.crawl.ISimpleScriptCrawl import ISimpleScriptCrawl

warnings.warn("'ISampleScriptCrawl'类已被废弃,请使用'ISimpleScriptCrawl'类代替 ", DeprecationWarning, 2)


class ISampleScriptCrawl(ISimpleScriptCrawl):
    """
    Python脚本式的Crawl的基类
    """

    @abstractmethod
    def do_crawl(self, params: dict):
        """
        抓取的业务逻辑
        :return: 返回一个 OutputWrapper的 数组
        """
        warnings.warn("'ISampleScriptCrawl'类已被废弃,请使用'ISimpleScriptCrawl'类代替 ", DeprecationWarning, 2)
        pass
